import os
import shutil
import tempfile
from pathlib import Path

import click

from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from ..decorators import loses_interactivity, require_login
from ..helpers.config import Config
from ..helpers.contexts import push_envs
from ..helpers.doctor import (
    AnsibleDoctorRunner,
    check_saml_client,
    deliver_logs,
    logs_dir,
    self_update,
    write_host_info,
    write_versions,
    write_versions_pip,
    zip_logs,
)
from ..helpers.global_options import GlobalOptions
from ..helpers.options import resource_argument
from .sym import sym


@sym.command(hidden=True, short_help="Run diagnostics")
@resource_argument
@click.option(
    "--mode",
    required=True,
    type=click.Choice(["basic", "ansible"]),
    default="basic",
    show_default=True,
    help="Whether or not to run Ansible checks",
)
@click.option(
    "--inventory",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="Path to an Ansible inventory to check",
)
@click.option(
    "--playbook",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="Path to an Ansible playbook to check",
)
@click.option(
    "--output-dir",
    required=True,
    default=str(Path.home() / "Desktop"),
    type=click.Path(file_okay=False, dir_okay=True, exists=True),
    show_default=True,
    help="Path to save output zip to",
)
@click.option(
    "--timeout",
    type=int,
    default=60,
    show_default=True,
    help="Max seconds to run any individual check",
)
@click.option("--did-update", is_flag=True, default=False, hidden=True)
@click.make_pass_decorator(GlobalOptions)
@require_login
@loses_interactivity
def doctor(
    options: GlobalOptions,
    resource: str,
    mode: str,
    playbook: str,
    inventory: str,
    output_dir: str,
    timeout: int,
    did_update: bool,
):
    """Run a series of diagnostic tests for Sym and output results to a zip file.

    RESOURCE is the name of a resource to check.
    """
    click.secho(
        f"Hey {Config.get_handle()}! Hope you're having a fantastic day :-)",
        fg="cyan",
        bold=True,
    )

    if not did_update:
        self_update()

    click.echo("Setting up test environment...")
    client = SAMLClientFactory.create_saml_client(resource, options)
    temp_dir = Path(tempfile.mkdtemp())
    doctor_dir = logs_dir(temp_dir)

    click.echo("Checking binary versions...")
    write_versions(doctor_dir / "versions.txt")

    click.echo("Checking Pip versions...")
    write_versions_pip(doctor_dir / "pip_versions.txt")

    click.echo("Checking SAML client...")
    check_saml_client(client, doctor_dir / "saml_client.txt")

    ssh_config_path = os.path.join(str(Path.home()), ".ssh", "config")
    if os.path.exists(ssh_config_path):
        click.echo("Copying SSH configuration...")
        shutil.copy(ssh_config_path, doctor_dir / "ssh_config")

    click.echo(f"Running {mode} checks...")
    with push_envs({"SYM_LOG_DIR": str(doctor_dir / "logs"), "SYM_DEBUG": "true"}):
        if mode == "ansible":
            AnsibleDoctorRunner(client, doctor_dir, timeout).run(inventory, playbook)

    click.echo("Checking host info...")
    write_host_info(doctor_dir / "host_info.txt")

    click.echo("Zipping it all up...")
    zip = zip_logs(doctor_dir)
    output = deliver_logs(client, Path(output_dir), zip)

    click.secho(f"Done! Your archive is at {str(output)}.", fg="green", bold=True)



from typing import List, Mapping, Sequence, Union

Options = Mapping[str, Union[str, bool]]
Argument = Union[str, Options]


def keywords_to_options(args: Sequence[Argument]) -> List[str]:
    """
    Converts any dicts in the given arguments into a sequence of
    options. The keys correspond to option names (with underscores
    replaced with hyphens). String values correspond to option
    arguments; boolean values correspond to the presence or
    absence of the option, with no argument.
    """

    result = []
    for arg in args:
        if isinstance(arg, str):
            result.append(arg)
            continue
        if arg is None:
            continue
        if isinstance(arg, list):
            result.extend(keywords_to_options(arg))
            continue
        for (key, value) in arg.items():
            if value in (False, None):
                continue
            prefix = "-" if len(key) == 1 else "--"
            option = prefix + key.replace("_", "-")
            if not isinstance(value, (tuple, list)):
                value = [value]
            for val in value:
                result.append(option)
                if isinstance(val, str):
                    result.append(val)
    return result



from sym.cli.helpers.os import find_command
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from ...errors import SAMLClientNotFound
from ..config import Config
from .model import CheckContext, CheckResult, SymCheck, failure, success


class LoginCheck(SymCheck):
    def check(self, _: CheckContext) -> CheckResult:
        if Config.is_logged_in():
            return success(f"Logged in as: {Config.get_email()} ({Config.get_org()})")
        else:
            return failure(f"Not logged in!")


class DependenciesCheck(SymCheck):
    def __init__(self, binary: str):
        self.binary = binary

    def check(self, _: CheckContext) -> CheckResult:
        if find_command(self.binary):
            return success(f"Found on path: {self.binary}")
        else:
            return failure(f"Could not find binary on path: {self.binary}")


class ResourceCheck(SymCheck):
    def check(self, ctx: CheckContext) -> CheckResult:
        try:
            client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
            msg = f"Created client with {client.binary} for resource {ctx.resource}"
            msg += "\nUsing config: " + str(client.config_file.path)
            msg += "\n" + client.config_file.read()

            return success(msg)
        except SAMLClientNotFound as e:
            return failure(
                f"Unable to create client: {ctx.options.saml_client_type} not found"
            )



import re

SentryDSN = "https://34647c66cad04a51b814cfade7b8d110@o373009.ingest.sentry.io/5342537"
SegmentWriteKey = "4RrqDnVYxMTE8wYJDNWKaPjHCSn9egeY"

AwsOktaNoRoles = re.compile("There are no roles that can be assumed|ARN isn't valid")
AwsOktaNotSetup = re.compile("404 Not Found")
AwsOktaNoCreds = re.compile("Okta credentials are not in your keyring")

Saml2AwsNoRoles = re.compile(
    "no (?:account|role)s available|Supplied RoleArn not found in saml assertion"
)
Saml2AwsNoCreds = re.compile("aws credentials not found")
Saml2AwsCredsExpired = re.compile("aws credentials have expired")



import pytest
from botocore.exceptions import ClientError
from botocore.paginate import Paginator

from sym.cli.errors import BotoError
from sym.cli.helpers.boto import send_ssh_key
from sym.cli.helpers.config import SymConfigFile
from sym.cli.helpers.ec2.client import Ec2Client
from sym.cli.helpers.ssh import SSHKeyPath


def test_error_handling_unauthorized(monkeypatch, boto_stub, click_context, saml_client):
    def paginate(*args, **kwargs):
        raise ClientError(
            {
                "Error": {
                    "Code": "UnauthorizedOperation",
                    "Message": "You are not authorized to perform this operation.",
                }
            },
            "DescribeInstances",
        )

    monkeypatch.setattr(Paginator, "paginate", paginate)

    with click_context:
        with pytest.raises(
            BotoError, match="Does your user role have permission to DescribeInstances?"
        ):
            Ec2Client(saml_client).load_instance_by_alias("localhost")


def test_error_handling_expired(monkeypatch, boto_stub, click_context, saml_client):
    def paginate(*args, **kwargs):
        raise ClientError(
            {
                "Error": {
                    "Code": "RequestExpired",
                    "Message": "Request has expired.",
                }
            },
            "DescribeInstances",
        )

    monkeypatch.setattr(Paginator, "paginate", paginate)

    with click_context:
        with pytest.raises(BotoError, match="Your AWS credentials have expired."):
            Ec2Client(saml_client).load_instance_by_alias("localhost")


def test_wait_for_invocation(saml_client, boto_stub, monkeypatch, click_context) -> None:
    send_command_response = {
        "Command": {
            "Status": "Pending",
            "CommandId": "1925e603-4f3f-4e1a-9f29-29fb7dcb4d47",
        }
    }
    get_command_invocation_response = {
        "Status": "Success",
        "CommandId": "1925e603-4f3f-4e1a-9f29-29fb7dcb4d47",
    }

    with click_context:
        ssm = boto_stub("ssm")
        ssm.add_response("send_command", send_command_response)
        ssm.add_response("get_command_invocation", get_command_invocation_response)

        ssh_key = SymConfigFile(file_name=f"{SSHKeyPath}.pub")
        ssh_key.put("==")

        send_ssh_key(saml_client, "foo", ssh_key)

        ssm.assert_no_pending_responses()



from multiprocessing import Pool

import pytest
from expects import *

from sym.cli.helpers.config import SymConfigFile
from sym.cli.saml_clients.saml2aws import Saml2Aws
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.saml_clients.conftest import TestContextFixture

pytestmark = [
    pytest.mark.usefixtures("click_context"),
    pytest.mark.parametrize(argnames=["constructor"], argvalues=[[Saml2Aws]]),
]


def test_saml2aws(
    test_context: TestContextFixture[Saml2Aws], capture_command: CaptureCommand
) -> None:
    with test_context(debug=False) as saml2aws:
        with capture_command():
            saml2aws.exec("aws", "ssm", "start-session", target="i-0123456789abcdef")
    capture_command.assert_command(
        f"saml2aws --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt exec -- env",
        f"saml2aws --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt login",
        f"saml2aws --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt exec -- 'aws ssm start-session --target i-0123456789abcdef'",
    )


def test_saml2aws_get_creds(
    test_context: TestContextFixture[Saml2Aws],
    capture_command: CaptureCommand,
) -> None:
    with test_context(debug=False) as saml2aws:
        with capture_command():
            expect(saml2aws.get_creds()).to(have_keys({"AWS_REGION": "foobar"}))


def test_saml2aws_debug(
    test_context: TestContextFixture[Saml2Aws], capture_command: CaptureCommand
) -> None:
    with test_context(debug=True) as saml2aws:
        with capture_command():
            saml2aws.exec("env")
    capture_command.assert_command(
        f"saml2aws --verbose --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt exec -- env",
        f"saml2aws --verbose --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt login",
        f"saml2aws --verbose --config {saml2aws.config_file} --idp-account sym-catamorphic --skip-prompt exec -- env",
    )


def test_saml2aws_multi_threads(subprocess_friendly_asserter) -> None:
    f, constructor, sandbox = subprocess_friendly_asserter
    with sandbox.push_xdg_config_home():
        config_file = SymConfigFile(resource="test", file_name="saml2aws.cfg")
    commands = (
        f"saml2aws --config {config_file} --idp-account sym-test --skip-prompt exec -- env",
        f"saml2aws --config {config_file} --idp-account sym-test --skip-prompt login",
        f"saml2aws --config {config_file} --idp-account sym-test --skip-prompt exec -- env",
    )
    with Pool(processes=4) as pool:
        pool.map(f, [(constructor, sandbox, commands)] * 100)



from click import BadParameter

from sym.cli.helpers.contexts import push_env
from sym.cli.sym import sym as click_command
from sym.cli.version import __version__


def test_version(click_setup):
    with click_setup(set_org=False, set_client=False) as runner:
        result = runner.invoke(click_command, ["version"])
        assert result.exit_code == 0
        assert result.output == f"{__version__}\n"


def test_login(click_setup):
    with click_setup(set_org=False) as runner:
        result = runner.invoke(
            click_command, ["login", "--org", "sym", "--email", "y@symops.io"]
        )
        assert result.exit_code == 0
        assert result.output == "Sym successfully initalized!\n"


def test_resources(click_setup):
    with click_setup() as runner:
        result = runner.invoke(click_command, ["resources"])
        assert result.exit_code == 0
        assert (
            result.output
            == "test (Test)\ntest_euw1 (Test: EU West 1)\ntest-custom (TestCustom)\n"
        )


def test_exec(click_setup, saml_client):
    with click_setup() as runner:
        result = runner.invoke(click_command, ["exec", "test", "--", "aws"])
        assert result.exit_code == 0


def test_env_vars(click_setup):
    with click_setup() as runner:
        with push_env("SYM_RESOURCE", "test"):
            result = runner.invoke(click_command, ["exec", "--", "aws"])
            assert result.exit_code == 0


def test_env_vars_invalid(click_setup):
    with click_setup() as runner:
        with push_env("SYM_RESOURCE", "tesst"):
            result = runner.invoke(click_command, ["exec", "--", "aws"])
            assert result.exit_code == BadParameter.exit_code
            assert "Invalid resource: tesst" in result.output


def test_env_vars_override(click_setup):
    with click_setup() as runner:
        with push_env("SYM_RESOURCE", "invalid"):
            result = runner.invoke(click_command, ["exec", "test", "--", "true"])
            assert result.exit_code == 0


def test_env_vars_override_invalid(click_setup):
    with click_setup() as runner:
        with push_env("SYM_RESOURCE", "invalid"):
            result = runner.invoke(click_command, ["exec", "invalid2", "--", "true"])
            assert result.exit_code == BadParameter.exit_code
            assert "Invalid resource: invalid" in result.output



from typing import Optional, Tuple

import click

from sym.cli.actions.action_registry import ActionRegistry
from sym.cli.data.ansible_options import AnsibleOptions
from sym.cli.data.request_data import RequestData
from sym.cli.decorators import command_require_bins, loses_interactivity, require_login
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.helpers.options import ansible_options as click_ansible_options
from sym.cli.helpers.options import resource_argument

from .sym import sym


@sym.command(
    short_help="Run an Ansible playbook",
    context_settings={"ignore_unknown_options": True},
)
@resource_argument
@click.argument("command", nargs=-1)
@click_ansible_options
@click.make_pass_decorator(GlobalOptions)
@command_require_bins("ansible-playbook", "aws", "session-manager-plugin")
@require_login
@loses_interactivity
def ansible_playbook(
    options: GlobalOptions,
    resource: str,
    command: Tuple[str, ...],
    ansible_aws_profile: Optional[str],
    ansible_sym_resource: Optional[str],
    control_master: bool,
    send_command: bool,
    forks: int,
) -> None:
    """Run Ansible commands against an inventory of EC2 instances, using
    approved credentials from the Sym RESOURCE group.

    For a list of available Sym RESOURCES, run `sym resources`.

    \b
    Example:
        `sym ansible-playbook RESOURCE -i path/to/inventory.txt path/to/playbook.yml`

    Extra arguments can be passed by adding them to the end of the command.

    \b
    Example:
        `sym ansible-playbook RESOURCE -i path/to/inventory.txt path/to/playbook.yml -vvvv -some -other -options`
    """

    ansible_options = AnsibleOptions(
        command=command,
        control_master=control_master,
        send_command=send_command,
        forks=forks,
        ansible_aws_profile=ansible_aws_profile,
        ansible_sym_resource=ansible_sym_resource,
    )

    request_data = RequestData(
        action="ansible_playbook",
        resource=resource,
        global_options=options,
        ansible_options=ansible_options,
    )
    ActionRegistry.execute(request_data)



from subprocess import CalledProcessError

from sym.cli.helpers.ec2.factory import get_ec2_client
from sym.cli.helpers.ssh import (
    ensure_public_key_on_instance,
    gen_local_ssh_config,
    maybe_gen_ssh_key,
    ssh_key_and_config,
)
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from .model import CheckContext, CheckResult, SymCheck, failure, success


class GenSshConfigCheck(SymCheck):
    def check(self, ctx: CheckContext) -> CheckResult:
        client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
        ssh_config = gen_local_ssh_config(client)
        msg = f"Successfully created local SSH config"
        msg += "\nUsing config: " + str(ssh_config.path)
        msg += "\n" + ssh_config.read()
        return success(msg)


class GenSshKeyCheck(SymCheck):
    def check(self, ctx: CheckContext) -> CheckResult:
        client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
        maybe_gen_ssh_key(client)
        ssh_key, _ = ssh_key_and_config(client)
        msg = "Local SSH keys generated"
        msg += "\nPrivate key: " + str(ssh_key.path)
        pub_key_path = str(ssh_key.path) + ".pub"
        msg += "\nPublic key: " + pub_key_path
        with open(pub_key_path) as f:
            msg += "\n" + f.read()
        return success(msg)


class EnsureSshKeyAuthorizedCheck(SymCheck):
    def __init__(self, instance: str):
        self.instance = instance

    def check(self, ctx: CheckContext) -> CheckResult:
        try:
            client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
            running_instance = get_ec2_client(client).load_instance_by_alias(
                self.instance
            )
            ensure_public_key_on_instance(client, running_instance.instance_id, 22)
            msg = f"Successfully ensured SSH key authorized on instance: {self.instance}"
            if self.instance != running_instance.instance_id:
                msg += f" [{running_instance.instance_id}]"
            return success(msg)
        except CalledProcessError as e:
            msg = f"Error checking SSH on instance {self.instance}: {e}"
            return failure(msg)



import json
import os
import re
import shlex
import shutil
import subprocess
import sys
import time
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple, cast

import click

from sym.cli.helpers.ansible import create_ssh_bin
from sym.cli.helpers.boto import put_file
from sym.cli.helpers.config import Config, SymConfigFile
from sym.cli.helpers.os import has_command
from sym.cli.helpers.params import get_ssh_user
from sym.cli.helpers.ssh import ssh_key_and_config
from sym.cli.saml_clients.saml_client import SAMLClient

BUCKET = "sym-doctor-prod-us-east-1"

BINARIES = [
    "aws",
    "ssh",
    "session-manager-plugin",
    "sym",
    "python",
    "python3",
    "pip",
    "pip3",
    "aws-okta",
    "saml2aws",
    "ansible",
]

PIP_PACKAGES = [
    "ansible",
    "boto",
    "boto3",
    "six",  # just here for a sanity check
]

REMOTE_LOG_DIRS = ["/var/log/amazon/ssm"]

TEST_PLAYBOOK = """
---
- name: Ping
  hosts: all
  become: true

  tasks:
    - ping:
"""

INSTANCE_PATTERN = re.compile(
    r"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|i-[a-f0-9]+)\s+:\sok="
)


def self_update():
    click.echo("Checking for updates...")
    version = self_update_pipx()
    if version is None:
        click.secho(
            "Failed to determine current version! It appears that sym is not a pipx-managed binary",
            fg="red",
        )
    else:
        old_version, new_version = version
        if old_version != new_version:
            click.secho(f"Updating to {new_version}!", fg="yellow")
            return subprocess.run([*sys.argv, "--did-update"])
        else:
            click.secho(f"{old_version} is the latest version!", fg="green")


def self_update_pipx() -> Optional[Tuple[str, str]]:
    if not has_command("pipx"):
        return None
    if not check_pipx_package_provides_binary("sym-cli", "sym"):
        return None
    version = get_version("sym")
    try:
        run_and_log_args(["pipx", "upgrade", "sym-cli"], check=True)
    except subprocess.CalledProcessError:
        return None
    return (cast(str, version), cast(str, get_version("sym")))


def check_saml_client(client, path: Path):
    message = f"client: {client.__class__.__name__}\nis_setup:{client.is_setup()}"
    click.secho(message, fg="blue")
    path.write_text(message)


def write_host_info(path: Path):
    path.write_text(run_and_log_args(["uname", "-a"]).stdout)


def get_version(binary) -> Optional[str]:
    if not has_command(binary):
        return None
    for command in ("--version", "-V", "version"):
        result = subprocess.run([binary, command], capture_output=True, text=True)
        if not result.returncode and (output := (result.stdout or result.stderr).strip()):
            return output


def check_pipx_package_provides_binary(package: str, binary: str):
    if (path := get_first_binary(binary)) :
        return os.path.join("pipx", "venvs", package, "bin", binary) in str(
            Path(path).expanduser().resolve()
        )
    else:
        return False


def get_version_pip(package):
    pip_binary = get_first_binary("pip", "pip3")
    packages = json.loads(run([pip_binary, "list", "--format", "json"]).stdout)
    return next((x for x in packages if x["name"] == package), {}).get("version")


def get_versions():
    versions = {}
    for binary in BINARIES:
        versions[binary] = get_version(binary)
        click.secho(f"Found {binary} with version {versions[binary]}", fg="blue")
    return versions


def get_versions_pip():
    pip_versions = {}
    for pip_package in PIP_PACKAGES:
        pip_versions[pip_package] = get_version_pip(pip_package)
        click.secho(
            f"Found Pip package {pip_package} with version {pip_versions[pip_package]}",
            fg="blue",
        )
    return pip_versions


def write_versions(path: Path):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text("\n".join(f"{k}\t\t{v}" for k, v in get_versions().items()))


def write_versions_pip(path: Path):
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text("\n".join(f"{k}\t\t{v}" for k, v in get_versions_pip().items()))


def logs_dir(temp_dir: Path) -> Path:
    return temp_dir / f"sym_logs_{datetime.now().date()}_{round(time.time())}"


def zip_logs(doctor_dir: Path) -> Path:
    zip = SymConfigFile(file_name=f"doctor/logs/{doctor_dir.name}")
    shutil.make_archive(str(zip.path), format="zip", root_dir=str(doctor_dir))
    return zip.path.with_suffix(".zip")


def deliver_logs(client: SAMLClient, output_dir: Path, zip: Path):
    output = output_dir / zip.name
    zip.link_to(output)

    try:
        put_file(
            client, BUCKET, zip, f"{Config.get_org()}/{Config.get_handle()}/{zip.name}"
        )
    except Exception as e:
        click.secho(f"Failed to upload logs: {e}", fg="red")

    return output


def run_and_log_args(args, **kwargs):
    return run(args, log=True, **kwargs)


def run(args, log=False, **kwargs):
    if log:
        click.secho(f"Running {shlex.join(args)}", fg="blue")
    return subprocess.run(args, capture_output=True, text=True, **kwargs)


def get_first_binary(*args):
    which_lines = run(["which", *args]).stdout.splitlines()
    return which_lines[0] if which_lines else None


def get_server_logs(client, doctor_dir: Path, instance: str):
    click.secho(f"Getting server logs from {instance}", fg="blue", bold=True)

    dest_dir = doctor_dir / "server_logs"
    dest_dir.mkdir(parents=True, exist_ok=True)

    ssh_user = get_ssh_user()
    ssh_bin = str(create_ssh_bin(client))
    ssh_key, _ = ssh_key_and_config(client)

    for dir in REMOTE_LOG_DIRS:
        ssh_commands = [
            f"mkdir -p /tmp/sym{dir}",
            f"sudo cp -r {dir}/ /tmp/sym{dir}",
            f"sudo chmod -R a+rw /tmp/sym{dir}",
        ]
        run_and_log_args(
            ["sym", "ssh", client.resource, instance, "; ".join(ssh_commands)]
        )

        run_and_log_args(
            [
                "scp",
                "-r",
                "-i",
                str(ssh_key),
                "-S",
                str(ssh_bin),
                f"{ssh_user}@{instance}:/tmp/sym{dir}",
                str(dest_dir),
            ]
        )

        ssh_commands = [
            f"cd /tmp/sym",
            f"rm -r /tmp/sym{dir}",
        ]
        run_and_log_args(
            ["sym", "ssh", client.resource, instance, "; ".join(ssh_commands)]
        )


def get_versions_from_server(client, doctor_dir: Path, instance: str):
    click.secho(f"Getting server version info from {instance}", fg="blue", bold=True)
    ssh_commands = [
        f"snap list amazon-ssm-agent",
        f"ssh -V 2>&1",
    ]
    res = run_and_log_args(
        ["sym", "ssh", client.resource, instance, "; ".join(ssh_commands)]
    )
    (doctor_dir / "server_versions.txt").write_text(res.stdout)


class DoctorRunner:
    def __init__(self, client: SAMLClient, doctor_dir: Path, timeout: int):
        self.client = client
        self.doctor_dir = doctor_dir
        self.timeout = timeout

    def run_with_timeout(self, args):
        return run_and_log_args(args, timeout=self.timeout)

    def write_error_logs(self, result, name: str = None):
        name = name or result.args[0]
        (self.doctor_dir / f"{name}_error_stdout.log").write_text(result.stdout)
        (self.doctor_dir / f"{name}_error_stderr.log").write_text(result.stderr)

    def write_temp_file(self, filename: str, content: str):
        path = self.doctor_dir.parent / filename
        path.write_text(content)
        return path


class AnsibleDoctorRunner(DoctorRunner):
    def run(self, inventory: Optional[str], playbook: Optional[str]):
        if not playbook:
            playbook = str(self.write_temp_file("playbook.yml", TEST_PLAYBOOK))

        args = ["sym", "ansible-playbook", self.client.resource]
        if inventory:
            args.extend(["-i", inventory])
        args.append(playbook)

        try:
            result = self.run_with_timeout(args)
        except subprocess.TimeoutExpired:
            click.secho("ansible check timed out", fg="red")
            return

        if (match := INSTANCE_PATTERN.search(result.stdout)) :
            get_server_logs(self.client, self.doctor_dir, match[1])
            get_versions_from_server(self.client, self.doctor_dir, match[1])
        else:
            self.write_error_logs(result, name="ansible_playbook")
            click.secho(f"Found no servers from which to retrieve logs", fg="red")



import re
import sys
from typing import Dict, Optional, Sequence

import click

from sym.cli.helpers.ec2.factory import get_ec2_client

from ..decorators import command_require_bins, loses_interactivity, require_login
from ..helpers.global_options import GlobalOptions
from ..helpers.keywords_to_options import keywords_to_options
from ..helpers.options import resource_argument
from ..helpers.ssh import raw_ssh, start_interactive_command, start_ssh_session
from ..saml_clients.saml_client_factory import SAMLClientFactory
from .sym import sym

SSH_MAN = """
     ssh [-AaCfGgKkMNnqsTtVvXxYy] [-B bind_interface] [-b bind_address] [-c cipher_spec]
         [-D [bind_address:]port] [-E log_file] [-e escape_char] [-F configfile] [-I pkcs11]
         [-i identity_file] [-J destination] [-L address] [-l login_name] [-m mac_spec] [-O ctl_cmd]
         [-o option] [-p port] [-Q query_option] [-R address] [-S ctl_path] [-W host:port]
         [-w local_tun[:remote_tun]] destination [command]
"""

SSH_OPTS = list(
    map(
        lambda x: x.casefold(),
        {
            "LogLevel",
            "CheckHostIP",
            "ForwardX11Trusted",
            "HostKeyAlias",
            "SetEnv",
            "UpdateHostKeys",
            "ClearAllForwardings",
            "GSSAPIKeyExchange",
            "HashKnownHosts",
            "remote_tun",
            "CertificateFile",
            "Hostname",
            "RemoteForward",
            "PubkeyAuthentication",
            "HostbasedKeyTypes",
            "AddKeysToAgent",
            "NumberOfPasswordPrompts",
            "KbdInteractiveDevices",
            "BatchMode",
            "PubkeyAcceptedKeyTypes",
            "CASignatureAlgorithms",
            "LocalCommand",
            "GSSAPIClientIdentity",
            "UserKnownHostsFile",
            "CanonicalizePermittedCNAMEs",
            "PermitLocalCommand",
            "HostName",
            "VerifyHostKeyDNS",
            "UsePrivilegedPort",
            "Port",
            "AddressFamily",
            "RemoteCommand",
            "HostKeyAlgorithms",
            "ForwardAgent",
            "CanonicalDomains",
            "VisualHostKey",
            "ExitOnForwardFailure",
            "EscapeChar",
            "GSSAPIDelegateCredentials",
            "IdentitiesOnly",
            "RekeyLimit",
            "TunnelDevice",
            "ControlMaster",
            "Ciphers",
            "KbdInteractiveAuthentication",
            "DynamicForward",
            "XAuthLocation",
            "SmartcardDevice",
            "ControlPath",
            "CanonicalizeHostname",
            "RequestTTY",
            "BindAddress",
            "StreamLocalBindUnlink",
            "StreamLocalBindMask",
            "RSAAuthentication",
            "RhostsRSAAuthentication",
            "ConnectionAttempts",
            "LocalForward",
            "PKCS11Provider",
            "PreferredAuthentications",
            "ServerAliveCountMax",
            "TCPKeepAlive",
            "Cipher",
            "local_tun",
            "Host",
            "ForwardX11Timeout",
            "Tunnel",
            "ChallengeResponseAuthentication",
            "GSSAPIKexAlgorithms",
            "HostbasedAuthentication",
            "CanonicalizeFallbackLocal",
            "Protocol",
            "ConnectTimeout",
            "ProxyJump",
            "SendEnv",
            "User",
            "GatewayPorts",
            "MACs",
            "Compression",
            "IPQoS",
            "GlobalKnownHostsFile",
            "ServerAliveInterval",
            "StrictHostKeyChecking",
            "KexAlgorithms",
            "CompressionLevel",
            "IdentityAgent",
            "ProxyUseFdpass",
            "GSSAPIAuthentication",
            "GSSAPIRenewalForcesRekey",
            "GSSAPITrustDns",
            "NoHostAuthenticationForLocalhost",
            "ProxyCommand",
            "PasswordAuthentication",
            "CanonicalizeMaxDots",
            "ControlPersist",
            "GSSAPIServerIdentity",
            "Match",
            "IdentityFile",
            "ForwardX11",
            "FingerprintHash",
        },
    )
)


def parse_ssh_man(ssh_man):
    flags_pattern = re.compile(r"ssh \[-(\w+)\]")
    options_pattern = re.compile(r"\[-(\w) (\S+)\]")

    flags = list(flags_pattern.search(ssh_man)[1])
    options = options_pattern.findall(ssh_man)

    return (flags, options)


def ssh_options(fn):
    flags, options = parse_ssh_man(SSH_MAN)
    for flag in flags:
        fn = click.option(f"-{flag}", flag, hidden=True, is_flag=True)(fn)
    for (option, name) in options:
        fn = click.option(
            f"-{option}", option, metavar=f"<{name}>", hidden=True, multiple=True
        )(fn)
    for opt in SSH_OPTS:
        fn = click.option(f"-o{opt}", opt, hidden=True, multiple=True)(fn)
    return fn


def normalize_token(token):
    if token.startswith("o") and len(token) > 1:
        return token.casefold()
    return token


def encode_options(kv):
    k, v = kv
    if k in SSH_OPTS:
        return {"o": [f"{k}={vv}" for vv in v]}
    else:
        return {k: v}


def use_interactive_ssm(is_a_tty: bool, command: Sequence[str], args: Dict) -> bool:
    if command:
        return False
    traditional_ssh_flags = ["D", "L", "N", "R", "T", "W", "w", "X", "Y"]
    for flag in traditional_ssh_flags:
        # if one of these flags is set to a truthy value then we need to speak traditional SSH
        if flag in args and args[flag]:
            return False
    return is_a_tty


@sym.command(
    short_help="Start a SSH session",
    context_settings={
        "ignore_unknown_options": True,
        "token_normalize_func": normalize_token,
    },
)
@resource_argument
@ssh_options
@click.option("-p", "--port", default=22, type=int, show_default=True)
@click.argument("destination", required=False)
@click.argument("command", nargs=-1, required=False)
@click.make_pass_decorator(GlobalOptions)
@command_require_bins("aws", "session-manager-plugin", "ssh")
@require_login
@loses_interactivity
def ssh(
    options: GlobalOptions,
    resource: str,
    destination: Optional[str],
    port: int,
    command: Sequence[str],
    **kwargs,
) -> None:
    """
    Run SSH into an instance within a Sym resource group, by host, IP or instance ID. Examples of usage:

    \b
    sym ssh prod i-abcdefgh1234
    sym ssh staging 10.20.30.40
    SYM_RESOURCE=staging sym ssh 10.20.30.40

    Standard SSH options can be supplied to the command like this:

    \b
    sym ssh test 10.20.30.40 -o ControlMaster=yes -o ControlPath=~/.tmp/cp -o ControlPersist=10s -v

    """
    ssh_args = keywords_to_options(map(encode_options, kwargs.items()))
    if not destination:
        """
        Ansible will use SSH commands with no destination to audit the local SSH capabilities. When
        we detect this is happening, pass the arguments through as-is.
        """
        raw_ssh({"p": str(port)}, *ssh_args)
        return

    client = SAMLClientFactory.create_saml_client(resource, options)

    client.dprint(f"ssh: args={ssh_args}")
    instance = get_ec2_client(client).load_instance_by_alias(destination)
    new_client = SAMLClientFactory.clone(client, aws_region=instance.region)

    interactive_ssm = use_interactive_ssm(
        is_a_tty=sys.stdin.isatty(),
        command=command,
        args=kwargs,
    )

    if interactive_ssm:
        start_interactive_command(new_client, instance.instance_id, command=command)
    else:
        start_ssh_session(
            new_client,
            instance.instance_id,
            port,
            args=ssh_args,
            command=command,
            wrap=False,
        )



import os
from contextlib import ExitStack, contextmanager
from typing import Iterator, Mapping, Optional

EnvValue = Optional[str]


def _set_env(key: str, value: EnvValue) -> EnvValue:
    last: EnvValue = os.getenv(key)

    if value is not None:
        os.environ[key] = value
    elif last is not None:
        del os.environ[key]

    return last


@contextmanager
def push_env(key: str, value: EnvValue) -> Iterator[EnvValue]:
    saved = _set_env(key, value)
    try:
        yield saved
    finally:
        _set_env(key, saved)


@contextmanager
def push_envs(envs: Mapping[str, EnvValue]) -> Iterator[EnvValue]:
    with ExitStack() as stack:
        for k, v in envs.items():
            stack.enter_context(push_env(k, v))
        yield



import click

from sym.cli.helpers.global_options import GlobalOptions

from ..actions.action_registry import ActionRegistry
from ..data.request_data import RequestData
from ..data.target_options import TargetOptions
from ..decorators import loses_interactivity, require_login, skip_analytics
from ..helpers.global_options import GlobalOptions
from ..helpers.options import resource_argument
from .sym import sym


@sym.command(hidden=True, short_help="Starts a SSH session over SSM")
@resource_argument
@click.option("--instance", help="Instance ID to connect to", required=True)
@click.option("--port", default=22, type=int, show_default=True)
@click.make_pass_decorator(GlobalOptions)
@skip_analytics
@require_login
@loses_interactivity
def ssh_session(options: GlobalOptions, resource: str, instance: str, port: int):
    target_options = TargetOptions(host=instance, port=port)
    request_data = RequestData(
        action="ssh_session",
        resource=resource,
        global_options=options,
        target_options=target_options,
    )

    ActionRegistry.execute(request_data)



# Env var names
ANSIBLE_SSH_ARGS = "ANSIBLE_SSH_ARGS"
ANSIBLE_DEBUG = "ANSIBLE_DEBUG"
ANSIBLE_SSH_EXECUTABLE = "ANSIBLE_SSH_EXECUTABLE"
ANSIBLE_SSH_RETRIES = "ANSIBLE_SSH_RETRIES"
ANSIBLE_PIPELINING = "ANSIBLE_PIPELINING"
SYM_USE_CONTROL_MASTER = "SYM_USE_CONTROL_MASTER"
SYM_DEBUG = "SYM_DEBUG"
SYM_LOG_DIR = "SYM_LOG_DIR"
OBJC_DISABLE_INITIALIZE_FORK_SAFETY = "OBJC_DISABLE_INITIALIZE_FORK_SAFETY"
ANSIBLE_CONNECTION_PLUGINS = "ANSIBLE_CONNECTION_PLUGINS"


# Env var values
DEFAULT_ANSIBLE_SSH_RETRIES = str(2)
ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER = "-C -o ControlMaster=no"
ANSIBLE_SSH_ARGS_DEBUG = "-o Compression=no -o ControlMaster=no"
DEFAULT_ANSIBLE_SSH_ARGS = "-C -o ControlMaster=auto -o ControlPersist=15m"
ENABLED = str(1)
YES = "YES"



from collections import defaultdict

import click
import immutables

from ..helpers.config import Config
from ..helpers.config.config import ConfigSchema
from ..helpers.validations import validate_resource
from .sym import sym

NO_VALIDATION = lambda _: True
VALIDATIONS = defaultdict(
    lambda _: NO_VALIDATION, {"default_resource": validate_resource}
)


def key_callback(ctx, key: str):
    if key not in ConfigSchema.__annotations__:
        raise click.BadParameter(f"Unrecognized configuration key `{key}`")
    return key


def value_callback(ctx, value: str):
    if value is None:
        return None
    if not VALIDATIONS[ctx.params["key"]](value):
        raise click.BadParameter(f"Invalid value `{value}` for key `{ctx.params['key']}`")
    return value


def pformat(value, indent=0):
    if isinstance(value, (dict, immutables.Map)):
        lines = []
        for k, v in value.items():
            tabs = "\t" * indent
            lines.append(f"\n{tabs}{k}: {pformat(v, indent=indent+1)}")
        return "\n".join(lines)
    else:
        return value


@sym.command(hidden=True, short_help="get a config value")
@click.argument("key", callback=key_callback)
@click.argument("value", callback=value_callback, required=False)
def config(key: str, value: str) -> None:
    """Get a Sym config value."""
    if value is None:
        click.echo(pformat(Config.instance().get(key)))
    else:
        Config.instance()[key] = value


@sym.command(hidden=True, name="config:set", short_help="set a config value")
@click.argument("key", callback=key_callback)
@click.argument("value", callback=value_callback)
def config_set(key: str, value: str) -> None:
    """Set a Sym config value."""
    Config.instance()[key] = value


@sym.command(hidden=True, name="config:unset", short_help="unset a config value")
@click.argument("key", callback=key_callback)
def config_unset(key: str) -> None:
    """Unset a Sym config value."""
    Config.instance()[key] = None



from getpass import getuser
from os import uname

from sentry_sdk.scope import Scope

from .config import Config


def set_scope_tag_org(scope: Scope) -> None:
    try:
        scope.set_tag("org", Config.get_org())
    except KeyError:
        pass


def set_scope_user(scope: Scope) -> None:
    user = {"username": getuser()}
    try:
        user["email"] = Config.get_email()
    except KeyError:
        pass
    scope.set_user(user)


def set_scope_context_os(scope: Scope) -> None:
    u = uname()
    uname_str = f"{u.sysname} {u.nodename} {u.release} {u.version} {u.machine}"
    scope.set_context(
        "os",
        {
            "name": u.sysname,
            "version": u.release,
            "build": u.version,
            "kernel_version": uname_str,
        },
    )



import os
from pathlib import Path
from typing import List

from sym.cli.actions.sym_action import SymAction
from sym.cli.data.request_data import RequestData
from sym.cli.helpers.check.core import DependenciesCheck, SymCheck
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.helpers.ssh import maybe_gen_ssh_key
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory


class WriteCredsAction(SymAction):
    def __init__(self):
        super().__init__("write_creds")

    def get_requirements(self) -> List[SymCheck]:
        return [DependenciesCheck("aws"), DependenciesCheck("session-manager-plugin")]

    def execute(self, request_data: RequestData) -> None:
        """Extract specifications from request_data and trigger execution
        to write credentials to an AWS credentials file.
        """

        self.write_creds(
            request_data.resource,
            request_data.get_global_options(),
            request_data.params.get("path"),
            request_data.params.get("profile"),
        )

    def write_creds(
        self, resource: str, options: GlobalOptions, path: Path, profile: str
    ) -> None:
        """Write out approved credentials for the specified resource to the profile
        in the AWS credentials file at the specified path.
        """

        client = SAMLClientFactory.create_saml_client(resource, options)
        client.write_creds(path=Path(path), profile=profile)
        maybe_gen_ssh_key(client)



import pytest

from sym.cli.helpers.ec2.client import InstanceNotFound
from sym.cli.helpers.ec2.multiregion import MultiRegionEc2Client
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.tests.helpers.ec2.conftest import (
    BOTO_EMPTY_RESPONSE,
    BOTO_INSTANCE_RESPONSE,
    TEST_INSTANCE,
    TEST_INSTANCE_ID,
    TEST_REGION,
)


@pytest.fixture
def multiregion_ec2_client(saml_client: SAMLClient) -> MultiRegionEc2Client:
    return MultiRegionEc2Client(saml_client)


def test_multiregion_find_in_other_region(
    ec2_stub, multiregion_ec2_client, click_context
):
    with click_context:
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
        )
        ec2_stub.add_response(
            "describe_regions",
            {"Regions": [{"RegionName": TEST_REGION}]},
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_INSTANCE_RESPONSE,
        )
        instance = multiregion_ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)
        assert instance == TEST_INSTANCE


def test_multiregion_find_in_default(ec2_stub, multiregion_ec2_client, click_context):
    with click_context:
        ec2_stub.add_response(
            "describe_instances",
            BOTO_INSTANCE_RESPONSE,
        )
        instance = multiregion_ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)
        assert instance == TEST_INSTANCE


def test_multiregion_not_found(ec2_stub, multiregion_ec2_client, click_context):
    with pytest.raises(InstanceNotFound, match=TEST_INSTANCE_ID):
        with click_context:
            ec2_stub.add_response(
                "describe_instances",
                BOTO_EMPTY_RESPONSE,
            )
            ec2_stub.add_response(
                "describe_regions",
                {"Regions": [{"RegionName": TEST_REGION}]},
            )
            ec2_stub.add_response(
                "describe_instances",
                BOTO_EMPTY_RESPONSE,
            )
            multiregion_ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)



import pytest

from sym.cli.decorators import retry


class NewError(Exception):
    pass


def test_retries():
    count = 0

    @retry(NewError)
    def fn():
        nonlocal count

        count += 1
        if count == 1:
            raise NewError

    fn()
    assert count == 2


def test_ultimately_raises():
    @retry(NewError)
    def fn():
        raise NewError

    with pytest.raises(NewError):
        fn()



from json import dumps

import click

from sym.cli.decorators import loses_interactivity, require_login, skip_analytics
from sym.cli.helpers.ec2.factory import get_ec2_client
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.helpers.options import resource_argument
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from .sym import sym


@sym.command(short_help="Get an Instance ID for a host")
@click.option("--json", is_flag=True, hidden=True)
@resource_argument
@click.argument("host")
@click.make_pass_decorator(GlobalOptions)
@skip_analytics
@require_login
@loses_interactivity
def host_to_instance(
    options: GlobalOptions,
    resource: str,
    host: str,
    json: bool,
) -> None:
    """Get the Instance ID of a HOST associated with a Sym RESOURCE."""
    client = SAMLClientFactory.create_saml_client(resource, options)
    instance = get_ec2_client(client).load_instance_by_alias(host)
    if json:
        click.echo(dumps({"instance": instance.instance_id, "region": instance.region}))
    else:
        click.echo(instance.instance_id)



import importlib.resources as pkg_resources
from pathlib import Path
from textwrap import dedent
from typing import Dict, Optional, Tuple

from sym.cli.constants.env import (
    ANSIBLE_CONNECTION_PLUGINS,
    ANSIBLE_DEBUG,
    ANSIBLE_PIPELINING,
    ANSIBLE_SSH_ARGS,
    ANSIBLE_SSH_ARGS_DEBUG,
    ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER,
    ANSIBLE_SSH_EXECUTABLE,
    ANSIBLE_SSH_RETRIES,
    DEFAULT_ANSIBLE_SSH_ARGS,
    DEFAULT_ANSIBLE_SSH_RETRIES,
    ENABLED,
    OBJC_DISABLE_INITIALIZE_FORK_SAFETY,
    YES,
)
from sym.cli.helpers.boto import get_identity
from sym.cli.helpers.config import SymConfigFile
from sym.cli.saml_clients.saml_client import SAMLClient

from ..ansible import connection
from ..helpers import os
from ..helpers.contexts import push_envs
from ..helpers.params import get_ansible_user
from ..helpers.ssh import ssh_key_and_config
from ..helpers.tee import Tee
from ..saml_clients.aws_profile import AwsCredentialsPath, AwsProfile
from ..version import __version__

ANSIBLE_SSH_PATH = "ansible/ssh"
ANSIBLE_SSH_PROFILE = "sym-ansible"

AWS_SSM_PATH = f"ansible/connection/{__version__}/sym_aws_ssm.py"

BUCKET_PREFIX = "sym-ansible-"
BUCKET_REGION = "us-east-1"


def get_ansible_bucket_name(client: SAMLClient) -> str:
    return client.get_profile().ansible_bucket or get_default_ansible_bucket_name(client)


def get_default_ansible_bucket_name(client: SAMLClient) -> str:
    account = get_identity(client)["Account"]
    return f"{BUCKET_PREFIX}{account}"


def aws_ssm_connection_plugin() -> Path:
    file = SymConfigFile(file_name=AWS_SSM_PATH, uid_scope=False)
    if not file.path.exists():
        file.put(pkg_resources.read_text(connection, file.path.name))
    return file.path


def ansible_connection_plugins_value() -> str:
    plugin = aws_ssm_connection_plugin()
    return ":".join(
        [
            str(plugin.parent),
            "~/.ansible/plugins/connection",  # these are the defaults
            "/usr/share/ansible/plugins/connection",
        ]
    )


def ansible_profile_proxy(client):
    return client.clone(klass=AwsProfile, resource=ANSIBLE_SSH_PROFILE)


def sym_subcommand(client, subcommand, args="") -> str:
    proxy_client = ansible_profile_proxy(client)
    return f"sym {proxy_client.cli_options} {subcommand} {proxy_client.resource} {args}"


def create_ssh_bin(client):
    proxy_client = ansible_profile_proxy(client)
    ssh_bin = proxy_client.subconfig(ANSIBLE_SSH_PATH)

    command = sym_subcommand(client, "ssh", '"$@"')
    if (log_dir := client.options.log_dir) :
        command = Tee.tee_command(log_dir, command)

    # fmt: off
    ssh_bin.put(dedent(
        f"""
        #!/bin/bash

        export PYTHONUNBUFFERED=1

        {command}
        """
    ).lstrip())
    # fmt: on
    ssh_bin.path.chmod(0o755)
    return ssh_bin


def get_ansible_env(
    ssh_bin: str,
    *,
    debug: bool = False,
    control_master: bool = True,
) -> Dict:
    envs = {
        ANSIBLE_SSH_EXECUTABLE: ssh_bin,
        ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
        ANSIBLE_PIPELINING: ENABLED,
    }
    if debug:
        envs[ANSIBLE_SSH_ARGS] = ANSIBLE_SSH_ARGS_DEBUG
        envs[ANSIBLE_DEBUG] = ENABLED
    elif not control_master:
        envs[ANSIBLE_SSH_ARGS] = ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER
    else:
        envs[ANSIBLE_SSH_ARGS] = DEFAULT_ANSIBLE_SSH_ARGS

    return envs


def get_send_command_args(client: SAMLClient):
    bucket_name = get_ansible_bucket_name(client)
    client.dprint(f"Using bucket s3://{bucket_name}")

    sym_host_to_instance_cmd = sym_subcommand(
        client,
        "host-to-instance",
        "$HOST --json",
    )
    extra_vars = {
        "ansible_aws_ssm_bucket": bucket_name,
        "ansible_aws_ssm_host_cmd": sym_host_to_instance_cmd,
        "ansible_aws_ssm_region": BUCKET_REGION,
        "ansible_aws_ssm_profile": ANSIBLE_SSH_PROFILE,
    }
    return " ".join(f"{k}='{v}'" for k, v in extra_vars.items())


def run_ansible(
    client: SAMLClient,
    command: Tuple[str, ...],
    *,
    ansible_aws_profile: Optional[str],
    ansible_sym_resource: Optional[str],
    forks: int,
    binary: str = "ansible",
    control_master: bool = True,
    send_command: bool = False,
):
    client.dprint(f"creating SSH artifacts")

    ssh_bin = str(create_ssh_bin(client))
    ssh_key, _ = ssh_key_and_config(client)
    args = [
        binary,
        f"--user={get_ansible_user()}",
        f"--private-key={ssh_key}",
        f"--scp-extra-args=-S '{ssh_bin}'",
        f"--sftp-extra-args=-S '{ssh_bin}'",
        *command,
    ]

    envs = get_ansible_env(ssh_bin, debug=client.debug, control_master=control_master)

    if send_command:
        envs[OBJC_DISABLE_INITIALIZE_FORK_SAFETY] = YES
        envs[ANSIBLE_CONNECTION_PLUGINS] = ansible_connection_plugins_value()

        args.append("--connection=sym_aws_ssm")
        args.append(f"--forks={forks}")
        args.append(f"--extra-vars={get_send_command_args(client)}")
    else:
        # Sometimes Ansible tries to use Paramiko instead of native SSH
        args.append("--connection=ssh")

    if client.debug:
        args.append("-vvvv")

    if client.log_dir:
        envs["ANSIBLE_LOG_PATH"] = str(Tee.path_for_fd(client.log_dir, "ansible"))
    client.write_creds(path=AwsCredentialsPath, profile=ANSIBLE_SSH_PROFILE)

    if ansible_aws_profile:
        ansible_client = client.clone(klass=AwsProfile, resource=ansible_aws_profile)
    elif ansible_sym_resource:
        ansible_client = client.clone(resource=ansible_sym_resource)
    else:
        ansible_client = client

    client.dprint(f"ansible: client='{ansible_client._cli_options()}', envs={envs}")

    with push_envs(envs):
        os.execvp(ansible_client, args)



from typing import Callable, Union

import click
from click_option_group import MutuallyExclusiveOptionGroup

from sym.cli.constants.env import SYM_USE_CONTROL_MASTER
from sym.cli.errors import CliError
from sym.cli.helpers.ssh import check_ssh_version
from sym.cli.helpers.util import flow

from ..saml_clients.chooser import choose_saml_client
from .config import Config
from .envvar_option import EnvvarGroupedOption, EnvvarOption
from .params import get_resource_env_vars


def config_option(
    name: str, help: str, default: Union[None, Callable[[], None]] = None, **kwargs
):
    def decorator(f):
        option_decorator = click.option(
            f"--{name}",
            help=help,
            prompt=True,
            default=default or (lambda: Config.instance().get(name)),
            **kwargs,
        )
        return option_decorator(f)

    return decorator


def _resource_arg_callback(ctx, _param, resource: str):
    """
    The resource *arg* callback raises an exception when the user isn't logged in,
    or when the supplied resource isn't valid for the logged in user.
    Otherwise, it returns the resource.
    """
    if not Config.is_logged_in():
        raise CliError("Please run `sym login` first")
    if resource is None:
        # We do some really complicated stuff in SymGroup to automagically update args so that
        # "resource" doesn't need to follow the strict rules of being a positional arg.
        # For "ansible-playbook" and "ssh" commands, this can result in this callback getting invoked
        # twice, once with "None" value here.
        return None
    saml_client = choose_saml_client(ctx.parent.params["saml_client_name"])
    if not resource or not saml_client.validate_resource(resource):
        raise click.BadParameter(f"Invalid resource: {resource}")
    return resource


def resource_option(f):
    """
    Used by the root command only. Subcommands that require a positional resource argument
    can get confused when the resource is provided via env var and is followed with arbitrary
    other args, so we have a special decorator in SymGroup copy the option into the
    right position. We don't use a callback or perform any special validation here,
    that validation will be performed when a subcommand uses the
    resource_argument decorator.
    """
    option_decorator = click.option(
        "--resource",
        help="The Sym resource to use",
        envvar=get_resource_env_vars(),
        default=lambda: Config.get_default("resource"),
        cls=EnvvarOption,
    )
    return option_decorator(f)


def resource_argument(f):
    option_decorator = click.argument(
        "resource",
        callback=_resource_arg_callback,
        envvar=get_resource_env_vars(),
        # We set required=False to ensure that the callback is always invoked. In the case where a
        # custom env var is used to specify the resource, but the user is not logged in, the callback
        # will correctly return a login error, rather than a missing argument error.
        required=False,
    )
    return option_decorator(f)


def ansible_options(f):
    group = MutuallyExclusiveOptionGroup("Ansible Roles")
    options = [
        group.option(
            "--ansible-aws-profile",
            help="The local AWS Profile to use for Ansible commands",
            envvar="AWS_PROFILE",
            cls=EnvvarGroupedOption,
        ),
        group.option(
            "--ansible-sym-resource",
            help="The Sym resource to use for Ansible commands",
            envvar="SYM_ANSIBLE_RESOURCE",
            callback=_resource_arg_callback,
            cls=EnvvarGroupedOption,
        ),
        click.option(
            "--control-master/--no-control-master",
            help="Allow SSH ControlPath caching",
            envvar=SYM_USE_CONTROL_MASTER,
            is_flag=True,
            default=check_ssh_version,
            cls=EnvvarOption,
        ),
        click.option(
            "--send-command/--no-send-command",
            help="Use SSM SendCommand instead of SSH",
            envvar="SYM_SEND_COMMAND",
            is_flag=True,
            default=True,
            cls=EnvvarOption,
            show_default=True,
        ),
        click.option(
            "--forks",
            help="Number of parallel subprocesses for ansible",
            default=10,
            show_default=True,
        ),
    ]
    return flow(options, f)


def required_option(*args, if_, **kwargs):
    def _callback(ctx, param: str, value: str):
        if not value and if_(ctx):
            raise click.exceptions.UsageError(f"{param} is required")
        return value

    def wrapper(f):
        return click.option(*args, callback=_callback, **kwargs)(f)

    return wrapper



import datetime
import os
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from semver import VersionInfo

V1 = VersionInfo.parse("0.0.1")


def test_pipx_path(pipx_updater):
    assert pipx_updater.is_pipx() == True


def test_brew_path(brew_updater):
    assert brew_updater.is_brew() == True


def test_local_if_v0(updater):
    updater.dir = Path("/foo/bar")
    assert updater.is_local() == True
    updater.get_current_version = MagicMock(return_value=V1)
    assert updater.is_local() == False


def test_default_path(updater):
    assert (updater.dir / "sym" / "cli").exists() == True


def test_noop_if_local(updater):
    assert updater.is_local() == True
    assert updater.should_check() == False


def test_no_manual_if_local(updater):
    updater.update = MagicMock()
    updater.manual_update()
    updater.update.assert_not_called()


def test_checks_first_time(pipx_updater):
    assert pipx_updater.should_check() == True


def test_no_check_second_time(pipx_updater):
    assert pipx_updater.should_check() == True
    assert pipx_updater.should_check() == False


def test_check_after_a_day(pipx_updater, time_travel):
    assert pipx_updater.should_check() == True
    time_travel.tick(delta=datetime.timedelta(days=1))
    assert pipx_updater.should_check() == True


def test_get_version(updater):
    assert updater.get_latest_version() is not None


def test_needs_update(pipx_updater):
    pipx_updater.get_current_version = MagicMock(return_value=V1)
    assert pipx_updater.needs_update() is True


def test_needs_no_update(pipx_updater):
    pipx_updater.get_current_version = MagicMock(return_value=V1)
    pipx_updater.get_latest_version = MagicMock(return_value=V1)
    assert pipx_updater.needs_update() is False


def test_failed_fetch(pipx_updater):
    pipx_updater.get_latest_version = MagicMock(return_value=None)
    assert pipx_updater.needs_update() is False


def test_update_pipx(pipx_updater, capture_command):
    with capture_command():
        assert pipx_updater._update() == True
    capture_command.assert_command(["pipx", "upgrade", "sym-cli"])


def test_update_pipx_failed(pipx_updater, capture_command):
    capture_command.register_output("pipx", "", exit_code=1)
    with capture_command():
        assert pipx_updater._update() == False


def test_update_brew(brew_updater, capture_command):
    with capture_command():
        assert brew_updater._update() == True
    capture_command.assert_command(["brew", "upgrade", "sym-cli"])


def test_update_brew_failed(brew_updater, capture_command):
    capture_command.register_output("brew", "", exit_code=1)
    with capture_command():
        assert brew_updater._update() == False


@pytest.mark.skipif(not os.getenv("CI"), reason="skipping real requests locally")
def test_update_other(other_updater):
    version = other_updater.get_latest_version()
    assert other_updater._update() == True
    assert (other_updater.dir / "sym").exists() == True
    assert str(version) in (other_updater.dir / "sym" / "cli" / "version.py").read_text()



import re
from subprocess import CalledProcessError

import pytest

from sym.cli.decorators import intercept_errors
from sym.cli.errors import CliError


class NewError(Exception):
    pass


@intercept_errors({re.compile("already"): NewError}, quiet=True)
def _raise_error(message: str) -> None:
    raise CalledProcessError(returncode=1, cmd=(), stderr=message)


def test_intercept_error_match(click_context) -> None:
    with pytest.raises(NewError):
        _raise_error("File already exists.")


def test_intercept_error_non_match(click_context) -> None:
    with pytest.raises(CliError):
        _raise_error("File not found.")



import importlib
from functools import reduce, wraps
from pathlib import Path
from typing import Any, List


def import_all(filename, name):
    for path in Path(filename).resolve().parent.glob("*.py"):
        if path.stem != "__init__":
            importlib.import_module(f".{path.stem}", name)


def requires_all_imports(import_all):
    def wrapper(fn):
        all_imported = False

        @wraps(fn)
        def wrapped(*args, **kwargs):
            nonlocal all_imported

            if not all_imported:
                import_all()
                all_imported = True

            return fn(*args, **kwargs)

        return wrapped

    return wrapper


def wrap(val):
    if isinstance(val, (list, tuple)):
        return val
    else:
        return (val,)


def flow(fns: List, val: Any):
    return reduce(lambda f, x: x(f), fns, val)



import os
import sys
from contextlib import contextmanager
from os import X_OK, access, get_exec_path, link
from os.path import join as join_path
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional, Sequence, TextIO

from portalocker import Lock
from portalocker.constants import LOCK_EX, LOCK_NB, LOCK_SH

from sym.cli.helpers.contexts import push_envs
from sym.cli.helpers.keywords_to_options import keywords_to_options


def read_lock(path: Path):
    return Lock(str(path), mode="r", flags=LOCK_SH | LOCK_NB)


# Many threads all contesting to create the _same_ file with the _same_ contents.
# This is a try-once-or-fail lock.
def create_lock(path: Path):
    return Lock(str(path), mode="w+", flags=LOCK_EX | LOCK_NB, fail_when_locked=True)


# Many threads contesting to atomically update shared state.
# Mode is r+ so that the buffer postion is at 0 for reading.
# This is effectively a spin lock.
def read_write_lock(path: Path):
    return Lock(
        str(path),
        mode="r+",
        flags=LOCK_EX | LOCK_NB,
        opener=lambda path, flags: os.open(path, flags | os.O_CREAT),
    )


@contextmanager
def safe_create(path: Path) -> TextIO:
    """
    Opens the given path in exclusive create mode (`"x"`), deleting the
    file beforehand if it already exists. This is the safe way to replace
    a file, that is not subject to symlink attacks.
    """
    path.parent.mkdir(parents=True, exist_ok=True)
    with NamedTemporaryFile(mode="w", dir=str(path.parent)) as f:
        yield f
        path.unlink(missing_ok=True)
        link(f.name, str(path))


@contextmanager
def cd(path: str):
    curr = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(curr)


def find_command(command: str, path: Optional[Sequence[str]] = None) -> str:
    try:
        return next(
            join_path(directory, command)
            for directory in (path or get_exec_path())
            if access(join_path(directory, command), X_OK)
        )
    except StopIteration:
        raise ValueError(command)


def has_command(command: str) -> bool:
    """
    Returns whether the given command can be found on `PATH` and is
    executable.
    """
    try:
        return bool(find_command(command))
    except ValueError:
        return False


def reset_fds():
    for fd in [sys.stdout, sys.stderr]:
        fd.flush()


def execvp(client, args: Sequence[str]):
    reset_fds()
    with push_envs(client.get_creds()):
        os.execvp(args[0], keywords_to_options(args))



import re
import socket
from dataclasses import dataclass
from typing import List, Sequence

import boto3
import validators
from botocore.exceptions import ClientError

from sym.cli.errors import CliError, InstanceNotFound
from sym.cli.helpers import boto

from ...saml_clients.saml_client import SAMLClient

InstanceIDPattern = re.compile("^i-[a-f0-9]+$")
UnauthorizedError = re.compile(r"UnauthorizedOperation")
RequestExpired = re.compile(r"RequestExpired")


@dataclass
class RunningInstance:
    instance_id: str
    region: str

    def __repr__(self) -> str:
        return f"{self.instance_id} [{self.region}]"

    def __eq__(self, other: "RunningInstance") -> bool:
        return self.instance_id == other.instance_id and self.region == other.region


class InstanceQuery:
    def __init__(self):
        self.instance_ids = []
        self.filters = [{"Name": "instance-state-name", "Values": ["running"]}]

    def instance_id(self, instance_id: str) -> "InstanceQuery":
        self.instance_ids.append(instance_id)
        return self

    def filter(self, filter_key: str, filter_values: List[str]) -> "InstanceQuery":
        self.filters.append({"Name": filter_key, "Values": filter_values})
        return self

    def params(self) -> dict:
        return {"Filters": self.filters, "InstanceIds": self.instance_ids}


class InvalidQueryError(CliError):
    def __init__(self, msg):
        super().__init__(msg)


class Ec2Client:
    """
    A basic client for EC2 that wraps the boto library and exposes methods for loading EC2 regions
    and instances. Note that instances are only loaded from the region defined in the SAMLClient
    passed into the constructor. To load instances from region "foo", make sure the client has been
    set up using the region "foo".
    """

    def __init__(self, client: SAMLClient):
        self.client = client

    @boto.intercept_boto_errors
    def get_regions(self) -> Sequence[str]:
        """
        List all the regions for the current session. This will first try to load the regions using the current
        credentials and executing `ec2 describe_regions`. If that call fails, then this function will fallback to
        all available regions for a default boto session.
        """
        try:
            client = boto.boto_client(self.client, "ec2")
            regions = [
                region["RegionName"] for region in client.describe_regions()["Regions"]
            ]
        except ClientError as e:
            self.client.dprint(
                "error describing regions, falling back to default available regions in boto",
                error=e,
            )
            session = boto3.Session()
            regions = session.get_available_regions("ec2")

        return sorted(regions, key=lambda r: ("us" not in r, "eu" not in r, r))

    def load_instance_by_alias(self, alias: str) -> RunningInstance:
        """
        Load instance by an alias, which could be an instance ID, IP address, or DNS name.

        If the instance can't be found, then raise InstanceNotFound(alias)
        """
        if InstanceIDPattern.match(alias):
            return self._load_instance(
                InstanceQuery().instance_id(alias),
                alias,
            )
        elif validators.ip_address.ipv4(alias):
            return self._load_instance_by_first_filter(
                ["ip-address", "private-ip-address"], alias
            )
        else:
            return self._load_instance_by_dns(alias)

    def _load_instance_by_dns(self, alias: str) -> RunningInstance:
        try:
            return self._load_instance_by_first_filter(
                ["dns-name", "private-dns-name"], alias
            )
        except InstanceNotFound as not_found:
            try:
                ip_addr = socket.gethostbyname(alias)
            except socket.error as e:
                self.client.dprint(
                    "Exception trying to resolve alias", alias=alias, error=e
                )
                raise not_found

            try:
                return self.load_instance_by_alias(ip_addr)
            except InstanceNotFound as e:
                self.client.dprint(
                    "Resolved alias into IP, but could not find instance by IP",
                    alias=alias,
                    ip_addr=ip_addr,
                )
                raise not_found

    @boto.intercept_boto_errors
    def _load_instances(self, query: InstanceQuery) -> List[RunningInstance]:
        """
        Load 0 or more instances by executing the supplied query. If the underlying
        call returns InvalidInstanceID.NotFound, because the query contains an unknown
        instance ID, then this function raises InstanceNotFound.
        """
        found = []
        client = boto.boto_client(self.client, "ec2")
        paginator = client.get_paginator("describe_instances")
        try:
            for response in paginator.paginate(
                Filters=query.filters,
                InstanceIds=query.instance_ids,
            ):
                for reservation in response["Reservations"]:
                    for instance in reservation["Instances"]:
                        found.append(
                            RunningInstance(
                                instance_id=instance["InstanceId"],
                                region=client.meta.region_name,
                            )
                        )
            return found
        except ClientError as e:
            if e.response["Error"]["Code"] == "InvalidInstanceID.NotFound":
                raise InstanceNotFound(query.instance_ids)
            raise

    def _load_instance(
        self, query: InstanceQuery, alias_for_errors: str
    ) -> RunningInstance:
        """
        Load exactly one instance by executing the supplied query.
        If the query returns 0 instances, this function raises InstanceNotFound.
        If the query returns 2 or more instances, this function raises InvalidQueryError.
        If the query raises an InstanceNotFound error, this function passes it.
        Otherwise, return the sole instance from the query result.
        """
        instances = self._load_instances(query)
        if len(instances) == 0:
            raise InstanceNotFound(alias_for_errors)
        elif len(instances) > 1:
            raise InvalidQueryError(
                f"Expected at most 1 instance matching {alias_for_errors}"
            )
        else:
            return instances[0]

    def _load_instance_by_first_filter(
        self, filter_keys: List[str], filter_value: str
    ) -> RunningInstance:
        """
        Load exactly one instance by searching against the set of filter_keys for the
        specified filter_value. The function stops when one of the filter key and value pairs
        matches an instance. If none match, this function raises TargetNotFound. If the underlying
        query returns InvalidQueryError, this function passes it.

        A common use case would be searching multiple fields ("ip-address", private-ip-address")
        for a particular IP address alias match. Returns immediately after finding a match in either key.
        """
        for key in filter_keys:
            try:
                return self._load_instance(
                    InstanceQuery().filter(key, [filter_value]),
                    filter_value,
                )
            except InstanceNotFound:
                pass

        raise InstanceNotFound(filter_value)



import os
import subprocess
import sys
import tarfile
from pathlib import Path
from subprocess import CalledProcessError, TimeoutExpired
from typing import Optional, Sequence
from urllib.error import URLError
from urllib.request import urlcleanup, urlretrieve

import click
import requests
from semver import VersionInfo

from ..version import __version__
from .config import Config
from .os import has_command, reset_fds

REQ_TIMEOUT = 1
INSTALL_TIMEOUT = 30


class SymUpdater:
    def __init__(self):
        self.dir = Path(__file__).resolve().parent.parent.parent.parent
        self.latest_version: Optional[VersionInfo] = None
        self.url: Optional[dict] = None

    def _log(self, s, **kwargs):
        click.secho(s, err=True, **kwargs)

    def should_check(self) -> bool:
        return not self.is_local() and Config.check_for_update()

    def get_current_version(self) -> VersionInfo:
        return VersionInfo.parse(__version__)

    def get_latest_version(self) -> Optional[VersionInfo]:
        try:
            resp = requests.get(
                "https://pypi.org/pypi/sym-cli/json",
                timeout=REQ_TIMEOUT,
            ).json()
            version = VersionInfo.parse(resp["info"]["version"])

            self.latest_version = version
            self.url = next(
                (u for u in resp["urls"] if u["python_version"] == "source"),
                None,
            )
            return version
        except (requests.exceptions.RequestException, ValueError, KeyError):
            return None

    def needs_update(self) -> bool:
        if (version := self.get_latest_version()) :
            return version > self.get_current_version()
        return False

    def is_local(self):
        return (self.dir / ".git").exists() or self.get_current_version() == "0.0.0"

    def is_pipx(self):
        return (Path.home() / ".local" / "pipx") in self.dir.parents

    def is_brew(self):
        return Path("/usr/local/Cellar") in self.dir.parents

    def _update_with_command(self, command: Sequence[str]) -> bool:
        if not has_command(command[0]):
            return False
        try:
            subprocess.run(
                command,
                timeout=INSTALL_TIMEOUT,
                check=True,
                capture_output=True,
            )
            return True
        except (CalledProcessError, TimeoutExpired):
            return False

    def update_pipx(self) -> bool:
        return self._update_with_command(["pipx", "upgrade", "sym-cli", "--force"])

    def update_brew(self) -> bool:
        return self._update_with_command(["brew", "upgrade", "sym-cli"])

    def update_other(self) -> bool:
        if not self.url:
            return False

        try:
            filename, _ = urlretrieve(self.url["url"])
        except URLError:
            urlcleanup()
            return False

        try:
            prefix = f"sym-cli-{self.latest_version}/"
            with tarfile.open(filename) as tar:
                for member in tar.getmembers():
                    if not member.name.startswith(prefix):
                        continue
                    member.name = str(Path(member.name).relative_to(prefix))
                    tar.extract(member, path=self.dir)
            return True
        except tarfile.TarError:
            return False

    def _update(self) -> bool:
        if self.is_pipx():
            return self.update_pipx()
        if self.is_brew():
            return self.update_brew()
        return self.update_other()

    def update(self, exec=False):
        self._log("Quickly checking for Sym CLI updates...", fg="white", nl=False)
        if not self.needs_update():
            self._log("no updates found, carry on!", fg="white")
            return
        self._log(f"found newer version {self.latest_version}!", fg="white", bold=True)

        self._log("Updating...", nl=False)
        if self._update():
            self._log("finished!", fg="green")
            if exec:
                reset_fds()
                os.execvp(sys.argv[0], sys.argv)
        else:
            self._log("failed :(", fg="red")

    def auto_update(self):
        try:
            if not self.should_check():
                return
            self.update(exec=True)
        except:
            return  # just to be safe

    def manual_update(self):
        if self.is_local():
            self._log("Auto-updating is not supported in dev environments!", fg="red")
            return
        self.update()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cli',
 'cli.actions',
 'cli.ansible',
 'cli.ansible.connection',
 'cli.commands',
 'cli.constants',
 'cli.data',
 'cli.helpers',
 'cli.helpers.check',
 'cli.helpers.config',
 'cli.helpers.ec2',
 'cli.saml_clients',
 'cli.tests',
 'cli.tests.commands',
 'cli.tests.decorators',
 'cli.tests.helpers',
 'cli.tests.helpers.ec2',
 'cli.tests.helpers.updater',
 'cli.tests.integration',
 'cli.tests.saml_clients']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'SecretStorage>=3.2.0,<4.0.0',
 'analytics-python>=1.2.9,<2.0.0',
 'boto3>=1.16.20,<2.0.0',
 'click-option-group>=0.5.1,<0.6.0',
 'click>=7.1.2,<8.0.0',
 'colorama>=0.4.4,<0.5.0',
 'immutables>=0.14,<0.15',
 'keyring>=21.5.0,<22.0.0',
 'policyuniverse>=1.3.2,<2.0.0',
 'portalocker>=2.0.0,<3.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'requests>=2.25,<3.0',
 'semver>=2.13.0,<3.0.0',
 'sentry-sdk>=0.19.3,<0.20.0',
 'validators>=0.18.1,<0.19.0']

entry_points = \
{'console_scripts': ['sym = sym.cli.sym:sym']}

setup_kwargs = {
    'name': 'sym-cli',
    'version': '0.0.69',
    'description': 'The CLI for Sym',
    'long_description': "# sym-cli\n\nSym CLI\n\n## Usage\n\n### Login\n\n`sym login --email rick@symops.io --org sym`\n\nOr, to run interactively: `sym login`\n\nCurrently, this saves the email and org name to a local configuration file that defaults to `$HOME/.config/sym/config.yml`.\nCurrently, this does NOT do any kind of validation or communication with the rest of the platform.\nThis location can be tweaked by editing the `XDG_CONFIG_HOME` environment variable (replaces `$HOME/.config`).\n\n### Resources\n\n`sym resources`\n\nOnce you run `sym login`, you can see the resource groups available by running `sym resources`.\n\nCurrently, this is based on a hard-coded list of resource groups defined in params.py. If logged in using a sym email address, then\nthere will be exactly one resource available: `test`.\n\n### SSH\n\n`sym ssh RESOURCE INSTANCE`\n\nInitializes an SSH connection to an instance in the resource group.\n\n`sym ssh test i-01073e9d0438334fc`\n\nCreates an SSH connection to an instance `i-01073e9d0438334fc`. In order to create this connection, Sym first attempts a login via SAML\nusing `saml2aws` (or `aws-okta` as a fallback) for the currently logged in user (the email address in `config.yml` set via `sym login`).\nThis may require MFA.\n\n`sym --saml-client=aws-profile ssh PROFILE INSTANCE`\n\n### Write Creds\n\n`sym write-creds test --profile PROFILE`\n\nIt can be convenient to save temporary credentials for a sym resource into an AWS profile. Then the AWS profile can be used for other commands\n(with `sym` or other tools).\n\n### Exec\n\n`sym exec test -- COMMAND`\n\nYou can use `sym` to execute a command or script using the credentials for the resource group.\n\nSimilar to other commands, you can use an AWS profile instead of a sym resource group:\n\n`sym --saml-client=aws-profile exec PROFILE -- COMMAND`\n\n### Ansible Playbook\n\n`sym ansible-playbook test -i ec2.py docker_test.yml`\n\nThere are several useful environment variables that can help with debugging:\n* `SYM_LOG_DIR`: Set to a directory (i.e. `/tmp/sym/test`) to accumulate logs from sym commands.\n* `SYM_DEBUG`: Set to `true` to turn on verbose logging from ansible and ssh\n\nLike other commands, you can use an AWS profile instead of a sym resource group:\n\n`sym --saml-client=aws-profile ansible-playbook PROFILE -i ec2.py docker_test.yml`\n\n#### Send Command\n\nAs of 0.0.45, the default implementation for Ansible to communicate with EC2 instances is via SSM using\nsend command. To do this, we ship an implementation of ansible's connection API in our Sym CLI, and that location\nis provided as a connection plugin argument to ansible. We provide options to that plugin implementation to\nmanage the connection and associated dependencies (s3 bucket for copying files over to the instance, etc).\n\nThis behavior can be enabled/disabled with a flag: `--send-command/--no-send-command`. `--send-command` is not required\nsince it is the default. Alternatively it can be toggled using the `SYM_SEND_COMMAND` environment variable.\n\n#### SSH\n\nWhen send command is _not_ used, then ansible falls back to the normal behavior - initiating an SSM session and then\nopening an SSH tunnel with that session. When using SSH with `sym ansible-playbook`, sym will force ansible to use\nopenssh instead of a built-in `Paramiko` implementation that is sometimes used, but was found to be buggy with SSM.\n\nWhen using SSH, ansible needs to decide whether to persist SSH sessions, which is important for performance, but on some\nversions of OpenSSH and aws-cli can be unreliable. The mechanism for persisting sessions is SSH control master\npersistence. Sym has the following behavior:\n* If the version of OpenSSH on the local machine is lower than a min version (currently 8.1+), then it will disable control master\nby default; otherwise, it will enable it by default.\n* Control master can be explicitly enabled/disabled with the `--control-master/--no-control-master` flag, or the `SYM_USE_CONTROL_MASTER` env var.\n* Control master (and compression) will be automatically disabled if `SYM_DEBUG` or `--debug` are set.\n\n### Doctor\n\n`sym doctor test --mode=ansible --inventory=test/integration/docker-ansible-ld/sym/ec2.py`\n\nUse `doctor` to run checks against the current environment and collect all the logs (sym, ansible,\nssh, etc.) from the local environment and remote instances.\n",
    'author': 'SymOps, Inc.',
    'author_email': 'pypi@symops.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/symopsio/sym-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import inspect
import logging
from dataclasses import dataclass, replace
from typing import Sequence


@dataclass
class GlobalOptionsBase:
    debug: bool = False

    def clone(self, **kwargs):
        return replace(self, **kwargs)

    def dprint(self, *s: Sequence[str], **kwargs):
        s = list(map(str, filter(None, s)))
        if (s or kwargs) and self.debug:
            message = " ".join(s)
            if kwargs:
                message += ": " + ",".join([f"{k}={v}" for k, v in kwargs.items()])
            mod = inspect.getmodule(inspect.stack()[1][0])
            logging.getLogger(mod.__name__).info(message)



import os
import tempfile
from contextlib import ExitStack
from typing import Dict, List

import pytest

from sym.cli.constants.env import (
    ANSIBLE_CONNECTION_PLUGINS,
    ANSIBLE_DEBUG,
    ANSIBLE_PIPELINING,
    ANSIBLE_SSH_ARGS,
    ANSIBLE_SSH_ARGS_DEBUG,
    ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER,
    ANSIBLE_SSH_RETRIES,
    DEFAULT_ANSIBLE_SSH_ARGS,
    DEFAULT_ANSIBLE_SSH_RETRIES,
    ENABLED,
    OBJC_DISABLE_INITIALIZE_FORK_SAFETY,
    SYM_DEBUG,
    SYM_LOG_DIR,
    SYM_USE_CONTROL_MASTER,
    YES,
)
from sym.cli.helpers.ansible import ansible_connection_plugins_value
from sym.cli.helpers.contexts import push_envs
from sym.cli.helpers.params import get_ssh_user
from sym.cli.tests.helpers.capture import CaptureCommand, c
from sym.cli.tests.helpers.sandbox import Sandbox

INVENTORY = "inventory.file"
PLAYBOOK = "playbook.file"
TEST_ACCOUNT = "foo-account"


_default_ansible_env = {
    ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
    ANSIBLE_SSH_ARGS: DEFAULT_ANSIBLE_SSH_ARGS,
    ANSIBLE_PIPELINING: ENABLED,
}


DEFAULT_SSH_VERSION_OUTPUT = "OpenSSH_8.1p1, LibreSSL 2.7.3"
OLD_SSH_VERSION_OUTPUT = "OpenSSH_7.1p1, LibreSSL 2.7.3"


def get_caller_identity_stub(make_stub):
    sts = make_stub("sts")
    sts.add_response("get_caller_identity", {"Account": TEST_ACCOUNT})


def _common_ansible_setup(
    capture_command: CaptureCommand,
    sandbox: Sandbox,
    ssh_output: str = DEFAULT_SSH_VERSION_OUTPUT,
):
    capture_command.register_output(
        "ssh",
        ssh_output,
        stderr=True,
    )
    sandbox.create_binary("bin/ansible-playbook")


def _assert_ansible_subcommands(
    capture_command: CaptureCommand,
    ansible_env: Dict[str, str] = _default_ansible_env,
):
    capture_command.assert_commands(
        [
            c("ssh", "-V"),
            c("exec", "true"),
            c(
                "ansible-playbook",
                "-i",
                INVENTORY,
                PLAYBOOK,
                "--connection=ssh",
                f"--user={get_ssh_user()}",
                **ansible_env,
            ),
        ]
    )


def _ansible_playbook_tester(command_tester, extra_args: List[str] = []):
    exit_stack = ExitStack()
    temp_dir = exit_stack.enter_context(tempfile.TemporaryDirectory())

    def setup():
        exit_stack.enter_context(push_envs({SYM_LOG_DIR: temp_dir}))

    def teardown():
        exit_stack.close()

    return command_tester(
        command=[
            "ansible-playbook",
            "test",
            "-i",
            INVENTORY,
            PLAYBOOK,
            *extra_args,
        ],
        setup=setup,
        teardown=teardown,
    )


@pytest.fixture
def ansible_playbook_tester_ssh(command_tester):
    return _ansible_playbook_tester(command_tester, ["--no-send-command"])


@pytest.fixture
def ansible_playbook_tester_ssh_no_control_master(command_tester):
    return _ansible_playbook_tester(
        command_tester, ["--no-send-command", "--no-control-master"]
    )


@pytest.fixture
def ansible_playbook_tester_send_command(command_tester):
    # TODO: we shouldn't need to pass this flag to bypass the SSH check when send-command is set
    return _ansible_playbook_tester(command_tester, ["--no-control-master"])


@pytest.fixture
def ansible_playbook_tester_send_command_low_forks(command_tester):
    # TODO: we shouldn't need to pass this flag to bypass the SSH check when send-command is set
    return _ansible_playbook_tester(command_tester, ["--no-control-master", "--forks=5"])


def test_ansible_playbook_ssh(
    ansible_playbook_tester_ssh,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup(_make_stub):
        _common_ansible_setup(capture_command, sandbox)

    with ansible_playbook_tester_ssh(setup=setup):
        _assert_ansible_subcommands(capture_command)


def test_ansible_playbook_ssh_old_version(
    ansible_playbook_tester_ssh,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup(_make_stub):
        _common_ansible_setup(
            capture_command,
            sandbox,
            ssh_output=OLD_SSH_VERSION_OUTPUT,
        )

    with ansible_playbook_tester_ssh(setup=setup):
        ansible_env = {
            ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
            ANSIBLE_SSH_ARGS: ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER,
            ANSIBLE_PIPELINING: ENABLED,
        }
        _assert_ansible_subcommands(capture_command, ansible_env)


def test_ansible_playbook_ssh_debug(
    ansible_playbook_tester_ssh,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    exit_stack = ExitStack()

    def setup(_make_stub):
        _common_ansible_setup(capture_command, sandbox)
        exit_stack.enter_context(push_envs({SYM_DEBUG: ENABLED}))

    def teardown():
        exit_stack.close()

    with ansible_playbook_tester_ssh(setup=setup, teardown=teardown):
        ansible_env = {
            ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
            ANSIBLE_SSH_ARGS: ANSIBLE_SSH_ARGS_DEBUG,
            ANSIBLE_DEBUG: ENABLED,
        }
        _assert_ansible_subcommands(capture_command, ansible_env)


def test_ansible_playbook_ssh_no_control_master_via_arg(
    ansible_playbook_tester_ssh_no_control_master,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup(_make_stub):
        _common_ansible_setup(capture_command, sandbox)

    with ansible_playbook_tester_ssh_no_control_master(setup=setup):
        ansible_env = {
            ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
            ANSIBLE_SSH_ARGS: ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER,
            ANSIBLE_PIPELINING: ENABLED,
        }
        # Since we're disabling control master with an arg, we skip the SSH check
        capture_command.assert_commands(
            [
                c("exec", "true"),
                c("ansible-playbook", "-i", INVENTORY, PLAYBOOK, **ansible_env),
            ]
        )


def test_ansible_playbook_ssh_no_control_master_via_env(
    ansible_playbook_tester_ssh,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    exit_stack = ExitStack()

    def setup(_make_stub):
        _common_ansible_setup(capture_command, sandbox)
        exit_stack.enter_context(push_envs({SYM_USE_CONTROL_MASTER: "false"}))

    def teardown():
        exit_stack.close()

    with ansible_playbook_tester_ssh(setup=setup, teardown=teardown):
        ansible_env = {
            ANSIBLE_SSH_RETRIES: DEFAULT_ANSIBLE_SSH_RETRIES,
            ANSIBLE_SSH_ARGS: ANSIBLE_SSH_ARGS_NO_CONTROL_MASTER,
            ANSIBLE_PIPELINING: ENABLED,
        }
        # Since we're disabling control master with an env var, we skip the SSH check
        capture_command.assert_commands(
            [
                c("exec", "true"),
                c("ansible-playbook", "-i", INVENTORY, PLAYBOOK, **ansible_env),
            ]
        )


def test_ansible_playbook_send_command(
    ansible_playbook_tester_send_command,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup(make_stub):
        get_caller_identity_stub(make_stub)
        sandbox.create_binary("bin/ansible-playbook")

    with ansible_playbook_tester_send_command(setup=setup):
        ansible_env = {
            ANSIBLE_PIPELINING: ENABLED,
            OBJC_DISABLE_INITIALIZE_FORK_SAFETY: YES,
            ANSIBLE_CONNECTION_PLUGINS: ansible_connection_plugins_value(),
        }
        ansible_args = [
            "--connection=sym_aws_ssm",
            "--forks=10",
            f"--user={get_ssh_user()}",
            f"--extra-vars=ansible_aws_ssm_bucket='sym-ansible-{TEST_ACCOUNT}' ansible_aws_ssm_host_cmd='sym --saml-client aws-profile --log-dir {os.environ['SYM_LOG_DIR']} host-to-instance sym-ansible $HOST --json' ansible_aws_ssm_region='us-east-1' ansible_aws_ssm_profile='sym-ansible'",
        ]
        capture_command.assert_commands(
            [
                c("exec", "true"),
                c(
                    "ansible-playbook",
                    "-i",
                    INVENTORY,
                    PLAYBOOK,
                    *ansible_args,
                    **ansible_env,
                ),
            ]
        )


def test_ansible_playbook_send_command_low_forks(
    ansible_playbook_tester_send_command_low_forks,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup(make_stub):
        get_caller_identity_stub(make_stub)
        sandbox.create_binary("bin/ansible-playbook")

    with ansible_playbook_tester_send_command_low_forks(setup=setup):
        ansible_env = {
            ANSIBLE_PIPELINING: ENABLED,
            OBJC_DISABLE_INITIALIZE_FORK_SAFETY: YES,
            ANSIBLE_CONNECTION_PLUGINS: ansible_connection_plugins_value(),
        }
        ansible_args = [
            "--connection=sym_aws_ssm",
            "--forks=5",
            f"--user={get_ssh_user()}",
            f"--extra-vars=ansible_aws_ssm_bucket='sym-ansible-{TEST_ACCOUNT}' ansible_aws_ssm_host_cmd='sym --saml-client aws-profile --log-dir {os.environ['SYM_LOG_DIR']} host-to-instance sym-ansible $HOST --json' ansible_aws_ssm_region='us-east-1' ansible_aws_ssm_profile='sym-ansible'",
        ]
        capture_command.assert_commands(
            [
                c("exec", "true"),
                c(
                    "ansible-playbook",
                    "-i",
                    INVENTORY,
                    PLAYBOOK,
                    *ansible_args,
                    **ansible_env,
                ),
            ]
        )


def test_ansible_no_login(command_login_tester):
    command_login_tester(["ansible"])
    command_login_tester(["ansible", "test"])
    command_login_tester(["ansible"], {"SYM_RESOURCE": "test"})
    command_login_tester(["ansible"], {"ENVIRONMENT": "test"})


def test_ansible_playbook_no_login(command_login_tester):
    command_login_tester(["ansible-playbook"])
    command_login_tester(["ansible-playbook", "test"])
    command_login_tester(
        ["ansible-playbook", "test", "-i", "inventory.txt", "playbook.yml"]
    )
    command_login_tester(["ansible-playbook"], {"SYM_RESOURCE": "test"})
    command_login_tester(["ansible-playbook"], {"ENVIRONMENT": "test"})
    command_login_tester(
        ["ansible-playbook", "-i", "inventory.txt", "playbook.yml"],
        {"ENVIRONMENT": "test"},
    )



import shlex

from sym.cli.helpers.keywords_to_options import Argument, keywords_to_options


def _assert_cmdline(expected: str, *args: Argument) -> None:
    assert keywords_to_options(args) == shlex.split(expected)


def test_no_args_pass_through() -> None:
    _assert_cmdline("")


def test_str_args_pass_through() -> None:
    _assert_cmdline("foo bar", "foo", "bar")


def test_mapping_args_with_str_values() -> None:
    _assert_cmdline("--foo bar --baz qux", {"foo": "bar", "baz": "qux"})


def test_mapping_args_with_bool_values() -> None:
    _assert_cmdline("--bar", {"foo": False, "bar": True})


def test_mix_and_match() -> None:
    _assert_cmdline(
        "saml2aws --verbose --idp-account sym exec aws ssm start-session --target i-0123456789abcdef",
        "saml2aws",
        {"verbose": True, "idp_account": "sym"},
        "exec",
        "aws",
        "ssm",
        "start-session",
        {"target": "i-0123456789abcdef"},
    )



import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, ContextManager, Iterator, Tuple, Type, TypeVar

import click
import pytest
from _pytest.monkeypatch import MonkeyPatch

from sym.cli.helpers.config import Config
from sym.cli.helpers.contexts import push_env
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.saml_clients import aws_profile
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.sym import sym as click_command
from sym.cli.tests.conftest import CustomOrgFixture
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.sandbox import Sandbox

P = TypeVar("P", bound=SAMLClient)
TestContextFixture = Callable[..., ContextManager[P]]


@pytest.fixture
def test_context(
    constructor: Type[P],
    sandbox: Sandbox,
    custom_org: CustomOrgFixture,
    capture_command: CaptureCommand,
    fake_creds_env_str: str,
) -> TestContextFixture[P]:
    @contextmanager
    def context(*, debug: bool) -> Iterator[P]:
        capture_command.register_output(r"env", fake_creds_env_str)
        with sandbox.push_xdg_config_home(), custom_org("launch-darkly"):
            sandbox.create_binary(f"bin/{constructor.binary}")
            with sandbox.push_exec_path():
                yield constructor("catamorphic", options=GlobalOptions(debug=debug))

    return context


@pytest.fixture
def test_context_with_creds(
    constructor: Type[P],
    sandbox: Sandbox,
    custom_org: CustomOrgFixture,
    monkeypatch: MonkeyPatch,
) -> TestContextFixture[P]:
    @contextmanager
    def context(*, debug: bool) -> Iterator[P]:
        path = sandbox.path / ".aws" / "credentials"
        creds = "\n".join(
            [
                "[catamorphic]",
                "aws_access_key_id=ASIA4GNNUMIHHBPSAQC3",
                "aws_secret_access_key=xxx",
                "aws_session_token=xxx",
            ]
        )
        monkeypatch.setattr(aws_profile, "get_identity", lambda x: {})
        with push_env("AWS_CREDENTIAL_FILE", str(path)):
            setattr(aws_profile, "AwsCredentialsPath", path)
            with sandbox.push_xdg_config_home(), custom_org("launch-darkly"):
                sandbox.create_binary(f"bin/{constructor.binary}")
                sandbox.create_file(".aws/credentials", 0o755, creds)
                with sandbox.push_exec_path():
                    yield constructor("catamorphic", options=GlobalOptions(debug=debug))

    return context


def _subprocess_friendly_asserter(args: Tuple[Type[P], Sandbox]) -> None:
    constructor, sandbox, expected_command = args
    monkeypatch = MonkeyPatch()
    capture_command = CaptureCommand(monkeypatch)

    monkeypatch.setattr(SAMLClient, "check_is_setup", lambda self: ...)

    with sandbox.push_xdg_config_home(), sandbox.push_exec_path(), capture_command():
        outputs = [
            "AWS_REGION=foobar\nAWS_FOOBAR=baz\nAWS_OKTA_SESSION_EXPIRATION=1600494616\n"
        ] * 2
        capture_command.enqueue_outputs(*outputs)
        Config.instance()["org"] = "sym"
        with click.Context(click_command) as ctx:
            ctx.ensure_object(GlobalOptions)
            constructor("test", options=GlobalOptions(debug=False)).exec("env")
            capture_command.assert_command(*expected_command)


@pytest.fixture
def subprocess_friendly_asserter(constructor: Type[P]):
    with tempfile.TemporaryDirectory() as tmpdirname:
        sandbox = Sandbox(Path(tmpdirname))
        sandbox.create_binary(f"bin/{constructor.binary}")
        yield (_subprocess_friendly_asserter, constructor, sandbox)



import re
import shlex
import sys
from abc import ABC, abstractmethod
from configparser import ConfigParser
from datetime import datetime, timedelta
from operator import attrgetter
from pathlib import Path
from typing import ClassVar, Optional, Sequence, Type
from urllib.parse import urlsplit

import dateutil.parser
from policyuniverse.arn import ARN

from sym.cli.errors import CliError
from sym.cli.helpers import segment, validations
from sym.cli.helpers.config import SymConfigFile
from sym.cli.helpers.keywords_to_options import keywords_to_options
from sym.cli.helpers.os import read_write_lock
from sym.cli.helpers.params import Profile, get_aws_saml_url, get_profile


class SAMLClient(ABC):
    binary: ClassVar[str]
    option_value: ClassVar[str]
    priority: ClassVar[int]
    setup_help: ClassVar[Optional[str]] = None

    resource: str
    options: "GlobalOptions"

    config_file: SymConfigFile

    _config: Optional[ConfigParser]
    _session_exists: bool
    _checked_setup: bool

    def __init__(self, resource: str, *, options: "GlobalOptions") -> None:
        self.resource = resource
        self.options = options

        self._config = None
        self._session_exists = False
        self._checked_setup = False

        self.check_is_setup()
        options.saml_clients[resource] = self

    @property
    def _section_name(self):
        return f"sym-{self.resource}"

    @property
    def debug(self):
        return self.options.debug

    @property
    def log_dir(self):
        return self.options.log_dir

    @classmethod
    def validate_resource(cls, resource: str):
        return validations.validate_resource(resource)

    def check_is_setup(self):
        if self._checked_setup:
            return
        self._checked_setup = True

        if self.is_setup():
            return

        print(
            f"Warning: sym might not function correctly until {self.binary} is setup.",
            file=sys.stderr,
        )
        if self.setup_help:
            print(f"Hint: {self.setup_help}", file=sys.stderr)

    @classmethod
    def sorted_subclasses(cls):
        filtered_subclasses = [
            c for c in cls.__subclasses__() if "Test" not in c.__name__
        ]
        return sorted(filtered_subclasses, key=attrgetter("priority"), reverse=True)

    @abstractmethod
    def _exec(self, *args: str, **opts: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def _ensure_config(self, profile: Profile) -> ConfigParser:
        raise NotImplementedError

    @abstractmethod
    def is_setup(self) -> bool:
        raise NotImplementedError

    def _cli_options(self):
        return {
            "saml_client": self.option_value,
            "debug": self.debug,
            "log_dir": self.log_dir,
        }

    @property
    def cli_options(self):
        return shlex.join(keywords_to_options([self._cli_options()]))

    def exec(self, *args: str, **kwargs: str) -> None:
        self.ensure_session()
        self.dprint(f"client: {self.cli_options}")
        return self._exec(*args, **kwargs)

    def _ensure_session(self, *, force: bool):
        self.dprint(f"ensuring session, force={force}")
        self._exec(
            "true",
            silence_stderr_=False,
            suppress_=True,
            run_subprocess_options_=self.options,
            intercept_errors_options_=self.options,
        )

    def ensure_session(self, force=False):
        if not self._session_exists:
            self._ensure_session(force=force or self._creds_expiring())
            self._session_exists = True

    def ensure_config(self) -> ConfigParser:
        if not self._config:
            config = self._ensure_config(self.get_profile())
            with self.config_file as f:
                config.write(f)
            self.dconfig(config)
            self._config = config
        return self._config

    def subconfig(self, file_name, **deps):
        return SymConfigFile(
            **self._cli_options(), resource=self.resource, **deps, file_name=file_name
        )

    def dprint(self, *s: Sequence[str], **kwargs):
        self.options.dprint(*s, **kwargs)

    def dconfig(self, config: ConfigParser):
        if self.debug:
            print("Writing config:", file=sys.stderr)
            config.write(sys.stderr)

    def log_subprocess_event(self, command: tuple):
        segment.track("Subprocess Called", global_options=self.options, binary=command[0])

    def get_profile(self) -> Profile:
        try:
            profile = get_profile(self.resource)
        except KeyError:
            raise CliError(f"Invalid resource: {self.resource}")

        self.dprint(f"Using profile {profile}")
        return profile

    def get_aws_saml_url(self, bare: bool = False) -> str:
        url = get_aws_saml_url(self.resource)
        if bare:
            url = urlsplit(url).path[1:]
        return url

    @property
    def _cred_env_vars(self):
        return (
            "AWS_REGION",
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_SESSION_TOKEN",
            "AWS_SECURITY_TOKEN",
            "AWS_CREDENTIAL_EXPIRATION",
        )

    def get_creds(self):
        output = self._exec(
            "env",
            capture_output_=True,
            censor_=True,
            run_subprocess_options_=self.options,
            intercept_errors_options_=self.options,
        )[-1]
        env_vars = re.findall(r"([\w_]+)=(.+)\n", output)
        return {k: v for k, v in env_vars if k in self._cred_env_vars}

    def _creds_expiring(self, buffer=timedelta(minutes=2)):
        try:
            creds = self.get_creds()
        except Exception:
            return True

        if (expires := creds.get("AWS_CREDENTIAL_EXPIRATION")) :
            try:
                parsed = dateutil.parser.isoparse(expires)
            except ValueError:
                return True  # Assume they are expiring as we're not sure
            remaining = parsed - datetime.now(parsed.tzinfo)
            self.dprint(f"{remaining} remaining in creds session")
            return remaining < buffer
        return False

    def _profile_matches_caller_identity(self):
        from sym.cli.helpers.boto import get_identity

        caller_identity = get_identity(self)
        caller_resource = self.resource_from_arn(caller_identity["Arn"])
        caller_role = caller_resource
        if "/" in caller_role:
            caller_role = caller_role.split("/", 2)[0]

        profile = self.get_profile()
        profile_role = self.resource_from_arn(profile.arn) if profile.arn else None

        self.dprint(f"*** Caller Role Name: {caller_role}")
        self.dprint(f"*** Profile Role Name: {profile_role}")

        if caller_role != profile_role:
            self.dprint(f"*** Caller identity does not match profile user!")
            return False

        return True

    def resource_from_arn(self, arn):
        if not arn:
            raise ValueError("SAMLClient missing ARN")

        name = ARN(arn).name
        if "/" in name:
            name = name.split("/", 1)[1]
        elif ":" in name:
            name = name.split(":", 1)[1]
        return name

    def write_creds(self, *, path: Path, profile: str):
        self.dprint(f"writing creds, profile={profile}")

        if (
            self._creds_expiring(buffer=timedelta(minutes=5))
            or not self._profile_matches_caller_identity()
        ):
            self._ensure_session(force=True)

        creds = self.get_creds()
        creds["region"] = creds.pop("AWS_REGION", self.get_profile().region)
        creds["x_security_token_expires"] = creds.pop("AWS_CREDENTIAL_EXPIRATION", None)
        creds["x_sym_ansible_bucket"] = self.get_profile().ansible_bucket

        path.parent.mkdir(parents=True, exist_ok=True)

        with read_write_lock(path) as f:
            config = ConfigParser()
            config.read_file(f)

            if config.has_section(profile):
                config.remove_section(profile)
            config.add_section(profile)
            for k, v in creds.items():
                if v:
                    config.set(profile, k.lower(), v)

            f.seek(0)
            f.truncate()
            config.write(f)

    def clone(self, *, klass: Type["SAMLClient"] = None, **overrides):
        kwargs = {
            key: overrides.get(key, getattr(self, key)) for key in ["resource", "options"]
        }
        return (klass or self.__class__)(**kwargs)



from pathlib import Path

from sym.cli.helpers.os import safe_create


def _assert_content(path: Path, expected: str) -> None:
    with path.open() as f:
        assert f.read() == expected


def test_safe_create(tmp_path: Path) -> None:
    foo_path = tmp_path / "foo"
    bar_path = tmp_path / "bar"
    with foo_path.open("x") as f:
        f.write("foo")

    bar_path.symlink_to("foo")
    _assert_content(bar_path, "foo")

    with safe_create(bar_path) as f:
        f.write("bar")
    _assert_content(bar_path, "bar")

    # This is, of course, what we're really testing for: that
    # foo didn't get overwritten.
    _assert_content(foo_path, "foo")



import shlex
import sys
from subprocess import CalledProcessError
from typing import Mapping, Pattern, Type, Union

from click import BadParameter, ClickException
from sentry_sdk import capture_exception

from .helpers.envvar_option import get_used

ErrorPatterns = Mapping[Pattern[str], Type[Exception]]


def raise_if_match(patterns: ErrorPatterns, msg: str):
    if error := next((err for pat, err in patterns.items() if pat.search(msg)), None):
        raise error


def get_active_env_vars():
    if used := get_used():
        mapping = "\n".join([f"{k}\t--{n}={v}" for k, (n, v) in used.items()])
        envvars = "\n".join(["Active Env Vars:", mapping])
        return f"\n\n{envvars}"
    else:
        return ""


def _format_message(self, file=None):
    return f"{self.message}{get_active_env_vars()}"


BadParameter.format_message = _format_message


class CliErrorMeta(type):
    _exit_code_count = 1

    def __new__(cls, name, bases, attrs):
        cls._exit_code_count += 1
        klass = super().__new__(cls, name, bases, attrs)
        klass.exit_code = cls._exit_code_count
        return klass


class CliError(ClickException, metaclass=CliErrorMeta):
    def __init__(self, *args, report=False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.report = report

    def show(self):
        if self.report:
            capture_exception()
        super().show()

    def format_message(self):
        return _format_message(self)


class CliErrorWithHint(CliError):
    def __init__(self, message, hint, **kwargs) -> None:
        msg = f"{message}\n\nHint: {hint}"
        super().__init__(msg, **kwargs)


class UnavailableResourceError(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "You don't have permission to access the Sym resource you requested.",
            "Request approval and then try again.",
        )


class ResourceNotSetup(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "The Sym resource you requested is not set up properly.",
            "Contact your Sym administrator.",
            report=True,
        )


class TargetNotConnected(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "The instance you tried to SSH into is not connected to AWS Session Manager.",
            "Ask your Sym administrator if they've set up Session Manager.",
        )


class AccessDenied(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "You don't have access to connect to this instance.",
            "Have you requested access?",
        )


class InstanceNotFound(CliError):
    def __init__(self, instance) -> None:
        super().__init__(f"Could not find instance {instance}")


class SAMLClientNotFound(CliError):
    def __init__(self) -> None:
        super().__init__(
            "No valid SAML client found in PATH. Supported clients are listed in `sym --help`."
        )


class OfflineError(CliError):
    def __init__(self) -> None:
        super().__init__("You are currently offline.")


class BotoError(CliErrorWithHint):
    def __init__(self, wrapped: "ClientError", hint: str) -> None:
        message = f"An AWS error occured!\n{str(wrapped)}"
        super().__init__(message, hint)


class FailedSubprocessError(CliError):
    def __init__(self, wrapped: Union[CalledProcessError, str]):
        if isinstance(wrapped, str):
            super().__init__(wrapped)
        else:
            super().__init__(f"Cannot run {shlex.join(wrapped.cmd)}")


class SuppressedError(FailedSubprocessError):
    def __init__(self, wrapped: CalledProcessError, echo=False):
        if echo:
            print(wrapped.stderr, file=sys.stderr)
        super().__init__(wrapped)

    def show(self):
        pass


class WrappedSubprocessError(FailedSubprocessError):
    def __init__(self, wrapped, hint, **kwargs) -> None:
        super().__init__(wrapped)
        messages = [
            f"Cannot run {wrapped.cmd[0]} [{wrapped.returncode}]",
            f"Hint: {hint}",
            f"\nThe original error was:",
            wrapped.stderr,
        ]
        CliError.__init__(self, "\n".join(messages), **kwargs)


class MissingPublicKey(WrappedSubprocessError):
    def __init__(self, wrapped, user) -> None:
        super().__init__(wrapped, f"Does the user ({user}) exist on the instance?")


class SamlClientNotSetup(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "Your SAML client is not set up.",
            "Consult the docs for your client.",
        )


class ExpiredCredentials(CliError):
    def __init__(self) -> None:
        super().__init__("Your AWS credentials have expired.")



from sym.cli.actions.sym_action import SymAction
from sym.cli.data.request_data import RequestData
from sym.cli.errors import CliError


class ActionRegistry:
    """
    Jumping off point to execute specific actions.  Note that overall this class has minimal
    dependencies.  It establishes a fixed strategy to:
        - Validate inputs
        - Load dependencies
        - Execute the referenced action
        - Log status and errors
    """

    @classmethod
    def create_ssh_action(cls) -> SymAction:
        from sym.cli.actions.ssh_session_action import SSHSessionAction

        return SSHSessionAction()

    @classmethod
    def create_write_creds_action(cls) -> SymAction:
        from sym.cli.actions.write_creds_action import WriteCredsAction

        return WriteCredsAction()

    @classmethod
    def create_ansible_action(cls) -> SymAction:
        from sym.cli.actions.ansible_action import AnsibleAction

        return AnsibleAction()

    @classmethod
    def create_ansible_playbook_action(cls) -> SymAction:
        from sym.cli.actions.ansible_playbook_action import AnsiblePlaybookAction

        return AnsiblePlaybookAction()

    @classmethod
    def instantiate_plugin(cls, action_name) -> SymAction:
        if action_name == "ssh_session":
            return cls.create_ssh_action()
        elif action_name == "write_creds":
            return cls.create_write_creds_action()
        elif action_name == "ansible":
            return cls.create_ansible_action()
        elif action_name == "ansible_playbook":
            return cls.create_ansible_playbook_action()

        raise CliError(f"unknown command: {action_name}")

    @classmethod
    def execute(cls, request_data: RequestData):
        """
        Note: This interface is a work in progress.  Right now it serves to call out to our action,
        but in the end I'd like to return an ExecutionScope object that does not exist.
        :param request_data:
        :return:
        """
        action = cls.instantiate_plugin(request_data.action)
        action.validate_inputs(request_data)
        action.invoke_requirements()
        return action.execute(request_data)



from typing import Optional

import click

from sym.cli.helpers.sym_group import SymGroup

from ..decorators import setup_segment, setup_sentry
from ..helpers.constants import SegmentWriteKey, SentryDSN
from ..helpers.envvar_option import EnvvarOption
from ..helpers.global_options import GlobalOptions
from ..helpers.options import resource_option
from ..helpers.sym_group import SymGroup
from ..helpers.updater import SymUpdater
from ..saml_clients.chooser import SAMLClientName, choose_saml_client, option_values
from ..version import __version__


@click.group(
    name="sym", cls=SymGroup, context_settings={"help_option_names": ["-h", "--help"]}
)
@click.option(
    "--saml-client",
    "saml_client_name",
    default="auto",
    type=click.Choice(option_values()),
    help="The SAML client type to use",
    envvar="SYM_SAML_CLIENT",
    cls=EnvvarOption,
)
@click.option(
    "--debug", is_flag=True, help="Enable verbose debugging", envvar="SYM_DEBUG"
)
@click.option("--aws-region", envvar="AWS_REGION", hidden=True)
@click.option(
    "--log-dir",
    type=click.Path(dir_okay=True, file_okay=False),
    hidden=True,
    envvar="SYM_LOG_DIR",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
@resource_option
@setup_segment(write_key=SegmentWriteKey)
@setup_sentry(dsn=SentryDSN, release=f"sym-cli@{__version__}")
def sym(
    options: GlobalOptions,
    saml_client_name: SAMLClientName,
    debug: bool,
    resource: str,
    aws_region: str,
    log_dir: Optional[str],
) -> None:
    """Access resources managed by Sym workflows.

    Use these tools to work with your resources once you've gotten approval in
    Slack.
    """
    options.saml_client_type = choose_saml_client(saml_client_name, none_ok=True)
    options.debug = debug
    options.log_dir = log_dir
    options.aws_region = aws_region

    SymUpdater().auto_update()



import json
import os
import subprocess
import tempfile

import pytest
from expects import *

from sym.cli.helpers.contexts import push_env
from sym.cli.helpers.os import cd

LD_DIR = "./test/integration/docker-ansible-ld/sym"

pytestmark = pytest.mark.skipif(
    not os.getenv("SYM_INTEGRATION"),
    reason="skipping integration tests",
)


def run_login(config, run):
    run(
        "login",
        "--email",
        config.getoption("email"),
        "--org",
        config.getoption("org"),
    )


def test_ssh_old_instance(pytestconfig, integration_runner):
    with integration_runner() as run:
        run_login(pytestconfig, run)
        result = run(
            "--saml-client",
            "aws-profile",
            "ssh",
            pytestconfig.getoption("resource"),
            pytestconfig.getoption("instance"),
            "--",
            "uname",
            "-a",
        )
        expect(result.output).to(contain("Ubuntu"))


def test_ansible_playbook_old_instance(pytestconfig, integration_runner):
    with cd(LD_DIR), integration_runner() as run:
        run_login(pytestconfig, run)
        with open("inventory.txt", "w") as f:
            f.write(pytestconfig.getoption("instance") + "\n")
        with tempfile.TemporaryDirectory() as log_dir:
            result = run(
                "--saml-client",
                "aws-profile",
                "--log-dir",
                log_dir,
                "ansible-playbook",
                pytestconfig.getoption("resource"),
                "-i",
                "inventory.txt",
                "docker_test.yml",
                "-vvv",
            )
        expect(result.output).to(contain("ok=5"))


@pytest.mark.skip(reason="Need to automate spinning up more instances")
def test_ansible_playbook_new_instance(pytestconfig, integration_runner):
    with cd(LD_DIR), integration_runner() as run:
        run_login(pytestconfig, run)
        with push_env("DO_RUN", "true"):
            subprocess.run(["./bin/ld_reserve_instance"], check=True)
        with tempfile.TemporaryDirectory() as log_dir:
            result = run(
                "--log-dir",
                log_dir,
                "ansible-playbook",
                pytestconfig.getoption("resource"),
                "-i",
                "ec2.py",
                "docker_test.yml",
            )
        expect(result.output).to(contain("ok=5"))


def test_exec(pytestconfig, integration_runner):
    with cd(LD_DIR), integration_runner() as run:
        run_login(pytestconfig, run)
        result = run(
            "--saml-client",
            "aws-profile",
            "exec",
            pytestconfig.getoption("resource"),
            "--",
            "./ec2.py",
        )
        expect(json.loads(result.output)).to(have_keys("_meta", "ec2"))



from typing import Dict

from sym.cli.data.ansible_options import AnsibleOptions
from sym.cli.data.target_options import TargetOptions
from sym.cli.helpers.global_options import GlobalOptions


class OptionsType:
    REQUEST = "Request"
    TARGET = "Target"
    ANSIBLE = "Ansible"


class RequestData:
    def __init__(
        self,
        action: str,
        resource: str,
        global_options: GlobalOptions = None,
        target_options: TargetOptions = None,
        ansible_options: AnsibleOptions = None,
        params: Dict = None,
    ):
        self.action = action
        self.resource = resource
        self.options = {}
        if params is None:
            self.params = {}
        else:
            self.params = params

        if global_options:
            self.set_options(OptionsType.REQUEST, global_options)

        if target_options:
            self.set_options(OptionsType.TARGET, target_options)

        if ansible_options:
            self.set_options(OptionsType.ANSIBLE, ansible_options)

    def get_options(self, key):
        return self.options.get(key, None)

    def set_options(self, key, value):
        self.options[key] = value

    def get_global_options(self):
        return self.get_options(OptionsType.REQUEST)

    def get_target_options(self):
        return self.get_options(OptionsType.TARGET)

    def get_ansible_options(self):
        return self.get_options(OptionsType.ANSIBLE)



from unittest.mock import patch

from expects import expect, match

from sym.cli.helpers.ansible import get_ansible_bucket_name
from sym.cli.helpers.params import Profile
from sym.cli.tests.commands.test_ansible import TEST_ACCOUNT, get_caller_identity_stub


def create_fake_profile_with_ansible_bucket(resource: str) -> Profile:
    """Mock function to return the sym organization's "test" AWS profile."""
    return Profile(
        display_name="Test",
        region="us-east-1",
        arn="arn:aws:iam::838419636750:role/SSMTestRole",
        ansible_bucket="foo",
    )


class TestAnsibleHelpers:
    """Test suite for ansible helper methods."""

    def test_get_default_ansible_bucket_name(self, boto_stub, saml_client):
        get_caller_identity_stub(boto_stub)
        actual = get_ansible_bucket_name(saml_client)
        expect(actual).to(match(f"sym-ansible-{TEST_ACCOUNT}"))

    @patch(
        "sym.cli.saml_clients.saml_client.get_profile",
        new=create_fake_profile_with_ansible_bucket,
    )
    def test_get_ansible_bucket_name_from_profile(self, boto_stub, saml_client):
        actual = get_ansible_bucket_name(saml_client)
        expect(actual).to(match("foo"))



import random
from contextlib import contextmanager

import pytest
from expects import *

from sym.cli.sym import sym as click_command
from sym.cli.tests.matchers import succeed

INSTANCES = [
    "i-052113589d5e8b832",
    "i-06ee7489efaeb582e",
    "i-0f3913172f1f4d1c0",
    "i-0ca07db4abb01d098",
    "i-0da5718079be3162c",
    "i-03371cb517e30eb22",
    "i-0e717dd52e0186713",
    "i-03c087811b23a2a4f",
    "i-08b3c84ebd6645df6",
    "i-0f5d63478e7cb781b",
]


@pytest.fixture
def integration_runner(capfdbinary, sandbox, wrapped_cli_runner):
    @contextmanager
    def context():
        runner = wrapped_cli_runner
        with sandbox.push_xdg_config_home():

            def run(*args):
                result = runner.invoke(click_command, args, catch_exceptions=False)
                cap = capfdbinary.readouterr()
                result.stdout_bytes = cap.out
                result.stderr_bytes = cap.err

                expect(result).to(succeed())
                return result

            yield run

    return context


def pytest_addoption(parser):
    parser.addoption("--email", default="ci@symops.io")
    parser.addoption("--org", default="sym")
    parser.addoption("--instance", default=random.choice(INSTANCES))
    parser.addoption("--resource", default="test")



import click

from ..decorators import require_login
from ..helpers.config import Config
from ..helpers.params import get_profiles
from .sym import sym


@sym.command(short_help="List available resource groups")
@require_login
def resources() -> None:
    """List resource groups available for use.

    Default resource is marked with an asterisk (*) if set.
    """
    default = Config.instance().get("default_resource")
    for (slug, profile) in get_profiles().items():
        star = "* " if default == slug else ""
        click.echo(f"{star}{slug} ({profile.display_name})")



import re
import sys
from configparser import ConfigParser
from datetime import datetime
from pathlib import Path
from typing import Final, Iterator, Optional, Tuple
from urllib.parse import urlsplit

import keyring

from ..decorators import intercept_errors, require_bins, run_subprocess
from ..errors import (
    FailedSubprocessError,
    ResourceNotSetup,
    SamlClientNotSetup,
    UnavailableResourceError,
)
from ..helpers.config import Config, SymConfigFile
from ..helpers.constants import AwsOktaNoCreds, AwsOktaNoRoles, AwsOktaNotSetup
from ..helpers.contexts import push_env
from ..helpers.keywords_to_options import Argument
from ..helpers.params import Profile, get_aws_okta_params
from .saml_client import SAMLClient

ErrorPatterns = {
    AwsOktaNoRoles: UnavailableResourceError,
    AwsOktaNotSetup: ResourceNotSetup,
    AwsOktaNoCreds: SamlClientNotSetup,
}

SessionIDPattern = re.compile(r"cache get `(.*)`")


class AwsOkta(SAMLClient):
    __slots__ = ["config_file", "resource", "options", "_config"]
    binary = "aws-okta"
    option_value = "aws-okta"
    priority = 5
    setup_help = "Run `aws-okta add`."

    resource: str
    options: "GlobalOptions"
    config_file: Final[SymConfigFile]
    _config: Optional[ConfigParser]

    def __init__(self, resource: str, *, options: "GlobalOptions") -> None:
        super().__init__(resource, options=options)
        self.config_file = SymConfigFile(resource=resource, file_name="aws-okta.cfg")

    def is_setup(self) -> bool:
        path = Path.home() / ".aws" / "config"
        config = ConfigParser()
        config.read(path)
        return "okta" in config.sections()

    @property
    def _cred_env_vars(self):
        return super()._cred_env_vars + ("AWS_OKTA_SESSION_EXPIRATION",)

    def get_creds(self):
        creds = super().get_creds()
        creds["AWS_CREDENTIAL_EXPIRATION"] = (
            datetime.fromtimestamp(int(creds["AWS_OKTA_SESSION_EXPIRATION"]))
            .astimezone()
            .replace(microsecond=0)
            .isoformat()
        )
        return creds

    @intercept_errors(ErrorPatterns)
    @run_subprocess
    @require_bins(binary)
    def _exec(
        self, *args: str, debug_: bool = False, **opts: str
    ) -> Iterator[Tuple[Argument, ...]]:
        self.log_subprocess_event(args)
        self.ensure_config()
        with push_env("AWS_CONFIG_FILE", str(self.config_file)):
            yield (
                "aws-okta",
                {"debug": self.debug or debug_},
                "exec",
                self._section_name,
                "--",
                *args,
                opts,
            )

    def _get_session_id(self):
        try:
            self._exec(
                "false",
                capture_output_=True,
                debug_=True,
                run_subprocess_options_=self.options,
                intercept_errors_options_=self.options,
            )
        except FailedSubprocessError as err:
            if match := SessionIDPattern.search(err.__cause__.stderr):
                return match[1]

    def _clear_session(self):
        if session_id := self._get_session_id():
            print(
                (
                    "Your credentials are about to expire."
                    "\nDue to a bug in aws-okta,"
                    " sym will now clear your cached credentials and request new ones."
                ),
                file=sys.stderr,
            )
            try:
                keyring.delete_password("aws-okta-login", session_id)
            except keyring.errors.KeyringError:
                pass

    @intercept_errors(ErrorPatterns)
    @run_subprocess
    @require_bins(binary)
    def _login(self) -> Iterator[Tuple[Argument, ...]]:
        self.ensure_config()
        with push_env("AWS_CONFIG_FILE", str(self.config_file)):
            yield (
                "aws-okta",
                {"debug": self.debug},
                "add",
                {"domain": self.okta_domain, "username": Config.get_email()},
                {
                    k: v
                    for k, v in get_aws_okta_params().items()
                    if k in ["mfa_provider", "mfa_factor_type"]
                },
            )

    def _ensure_session(self, *, force: bool):
        if force:
            self._clear_session()
        # aws-okta fails in mysterious ways if your
        # keychain credentials are deleted after setup.
        # This logic handles that case cleanly.
        try:
            super()._ensure_session(force=force)
        except FailedSubprocessError:
            print("Logging in to Okta...", file=sys.stderr)
            self._login(silence_stderr_=False)
            print("Fetching SAML assertion...", file=sys.stderr)
            super()._ensure_session(force=force)

    @property
    def okta_domain(self) -> str:
        url = self.get_aws_saml_url(bare=False)
        return urlsplit(url).netloc

    def _ensure_config(self, profile: Profile) -> ConfigParser:
        config = ConfigParser(default_section="okta")
        config.read_dict(
            {
                "okta": get_aws_okta_params(),
                f"profile {self._section_name}": {
                    "aws_saml_url": self.get_aws_saml_url(bare=True),
                    "region": profile.region,
                    "role_arn": profile.arn,
                    **get_aws_okta_params(),
                },
            }
        )
        return config



import json
import re
import shlex
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from textwrap import dedent
from typing import Dict, Sequence

import boto3
from botocore.config import Config as BotoConfig
from botocore.exceptions import ClientError, ConnectionError, ProfileNotFound, WaiterError

from ..decorators import retry
from ..errors import BotoError, CliError, OfflineError
from .config import SymConfigFile
from .params import get_ssh_user

InstanceIDPattern = re.compile("^i-[a-f0-9]+$")
UnauthorizedError = re.compile(r"UnauthorizedOperation")
RequestExpired = re.compile(r"RequestExpired|ExpiredTokenException")


def intercept_boto_errors(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ConnectionError as err:
            raise OfflineError() from err
        except ClientError as err:
            if UnauthorizedError.search(str(err)):
                raise BotoError(
                    err, f"Does your user role have permission to {err.operation_name}?"
                )
            if RequestExpired.search(str(err)):
                raise BotoError(
                    err,
                    f"Your AWS credentials have expired. Try running `sym write-creds` again.",
                )

            raise CliError(str(err)) from err
        except ProfileNotFound as err:
            raise BotoError(
                err,
                "AWS profile specified could not be found. "
                "Please check your `AWS_PROFILE` environment variable and "
                "your credentials file (default: $HOME/.aws/credentials).",
            )

    return wrapped


def boto_client(saml_client, service):
    creds = saml_client.get_creds()
    return boto3.client(
        service,
        config=BotoConfig(region_name=get_region(saml_client, creds)),
        aws_access_key_id=creds["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=creds["AWS_SECRET_ACCESS_KEY"],
        aws_session_token=creds.get("AWS_SESSION_TOKEN"),
    )


def get_region(saml_client, creds=None) -> str:
    override = saml_client.options.aws_region
    return override or (creds or saml_client.get_creds()).get("AWS_REGION")


@intercept_boto_errors
def send_ssh_key(saml_client: "SAMLClient", instance: str, ssh_key: SymConfigFile):
    saml_client.dprint("sending SSH key")

    user = get_ssh_user()
    ssm_client = boto_client(saml_client, "ssm")
    # fmt: off
    command = dedent(
        f"""
        #!/bin/bash
        mkdir -p "$(echo ~{user})/.ssh"
        echo "{ssh_key.path.with_suffix('.pub').read_text()}" >> "$(echo ~{user})/.ssh/authorized_keys"
        chown -R {user}:{user} "$(echo ~{user})/.ssh"
        """
    ).strip()
    # fmt: on

    response = ssm_client.send_command(
        InstanceIds=[instance],
        DocumentName="AWS-RunShellScript",
        Comment="SSH Key for Sym",
        Parameters={"commands": command.splitlines()},
    )

    saml_client.dprint(response)

    _wait_for_command(ssm_client, instance, response)


@retry(
    WaiterError,
    count=5,
    delay=1,
    check_ex=lambda ex: ex.last_response["Error"]["Code"] == "InvocationDoesNotExist",
)
def _wait_for_command(ssm_client, instance, response):
    waiter = ssm_client.get_waiter("command_executed")
    waiter.wait(
        InstanceId=instance,
        CommandId=response["Command"]["CommandId"],
        WaiterConfig={"Delay": 1},
    )


@intercept_boto_errors
def get_identity(saml_client) -> dict:
    sts_client = boto_client(saml_client, "sts")
    return sts_client.get_caller_identity()


@intercept_boto_errors
def _ssm_command(saml_client, ssm_client, params: Dict):
    saml_client.dprint("SSM Session: ", params=params)
    response = ssm_client.start_session(**params)
    command = [
        "session-manager-plugin",
        json.dumps(response),
        ssm_client.meta.region_name,
        "StartSession",
        "",  # This is the profile name, left empty since we're supplying credentials via environment
        json.dumps(params),
        ssm_client.meta.endpoint_url,
    ]
    return response, command


@contextmanager
def _ssm_session(saml_client, params: Dict):
    ssm_client = boto_client(saml_client, "ssm")
    response, command = _ssm_command(saml_client, ssm_client, params=params)
    try:
        yield command
    finally:
        try:
            ssm_client.terminate_session(SessionId=response["SessionId"])
        except ClientError:
            pass  # Creds expired while session was in progress, cannot terminate it


@contextmanager
def ssm_ssh_session(saml_client, instance: str, port: int):
    params = {
        "Target": instance,
        "DocumentName": "AWS-StartSSHSession",
        "Parameters": {"portNumber": [str(port)]},
    }
    with _ssm_session(saml_client, params) as command:
        yield command


@contextmanager
def ssm_interactive_command(saml_client, instance: str, user: str, cmd: Sequence[str]):
    full_cmd = f"sudo run-parts /etc/update-motd.d; sudo su -l {user}"
    if len(cmd) > 0:
        joined = shlex.join(cmd)
        full_cmd += f' -c "{joined}"'

    params = {
        "Target": instance,
        "DocumentName": "AWS-StartInteractiveCommand",
        "Parameters": {"command": [f"{full_cmd}"]},
    }
    with _ssm_session(saml_client, params) as command:
        yield command


@intercept_boto_errors
def put_file(saml_client, bucket: str, file_path: Path, object_path: str):
    s3_client = boto_client(saml_client, "s3")
    s3_client.upload_file(
        str(file_path),
        bucket,
        object_path,
        ExtraArgs={"ACL": "bucket-owner-full-control"},
    )



import pytest

from sym.cli.helpers.ec2.client import Ec2Client, RunningInstance
from sym.cli.saml_clients.saml_client import SAMLClient

TEST_REGION = "test-region-1"
TEST_INSTANCE_ID = "i-0123456789abcdef"
TEST_IP = "10.20.30.40"
TEST_DNS_NAME = "ip-10-20-30-40.ec2.internal"
TEST_HOSTNAME = "test-host.com"
DEFAULT_REGION = "us-east-1"
TEST_INSTANCE = RunningInstance(instance_id=TEST_INSTANCE_ID, region=DEFAULT_REGION)

BOTO_EMPTY_RESPONSE = {"Reservations": []}
BOTO_INSTANCE_RESPONSE = {
    "Reservations": [
        {
            "Instances": [
                {
                    "InstanceId": TEST_INSTANCE_ID,
                }
            ]
        }
    ]
}


@pytest.fixture
def ec2_stub(boto_stub):
    return boto_stub("ec2")


@pytest.fixture
def ec2_client(saml_client: SAMLClient) -> Ec2Client:
    return Ec2Client(saml_client)


@pytest.fixture
def region(saml_client: SAMLClient) -> str:
    return saml_client.get_profile().region



from typing import List

import click

from sym.cli.helpers.check.runner import must_run_all

from ..decorators import loses_interactivity
from ..helpers.global_options import GlobalOptions
from .sym import sym


@sym.command(hidden=True, short_help="Runs a series of checks against a Sym resource")
@click.make_pass_decorator(GlobalOptions)
@click.argument("resource")
@click.option("--instance", help="Instance ID(s) to check", multiple=True, default=[])
@loses_interactivity
def check(options: GlobalOptions, resource: str, instance: List[str]):
    """Run a series of checks against a Sym RESOURCE."""
    must_run_all(options, resource, instance)
